# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for ledger_api protocol."""

from typing import Any, Dict, cast

from aea.protocols.base import Message, Serializer

from packages.fetchai.protocols.ledger_api import ledger_api_pb2
from packages.fetchai.protocols.ledger_api.custom_types import (
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage


class LedgerApiSerializer(Serializer):
    """Serialization for the 'ledger_api' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'LedgerApi' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(LedgerApiMessage, msg)
        ledger_api_msg = ledger_api_pb2.LedgerApiMessage()
        ledger_api_msg.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        ledger_api_msg.dialogue_starter_reference = dialogue_reference[0]
        ledger_api_msg.dialogue_responder_reference = dialogue_reference[1]
        ledger_api_msg.target = msg.target

        performative_id = msg.performative
        if performative_id == LedgerApiMessage.Performative.GET_BALANCE:
            performative = ledger_api_pb2.LedgerApiMessage.Get_Balance_Performative()  # type: ignore
            ledger_id = msg.ledger_id
            performative.ledger_id = ledger_id
            address = msg.address
            performative.address = address
            ledger_api_msg.get_balance.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.GET_RAW_TRANSACTION:
            performative = ledger_api_pb2.LedgerApiMessage.Get_Raw_Transaction_Performative()  # type: ignore
            terms = msg.terms
            Terms.encode(performative.terms, terms)
            ledger_api_msg.get_raw_transaction.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION:
            performative = ledger_api_pb2.LedgerApiMessage.Send_Signed_Transaction_Performative()  # type: ignore
            signed_transaction = msg.signed_transaction
            SignedTransaction.encode(
                performative.signed_transaction, signed_transaction
            )
            ledger_api_msg.send_signed_transaction.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT:
            performative = ledger_api_pb2.LedgerApiMessage.Get_Transaction_Receipt_Performative()  # type: ignore
            transaction_digest = msg.transaction_digest
            TransactionDigest.encode(
                performative.transaction_digest, transaction_digest
            )
            ledger_api_msg.get_transaction_receipt.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.BALANCE:
            performative = ledger_api_pb2.LedgerApiMessage.Balance_Performative()  # type: ignore
            ledger_id = msg.ledger_id
            performative.ledger_id = ledger_id
            balance = msg.balance
            performative.balance = balance
            ledger_api_msg.balance.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.RAW_TRANSACTION:
            performative = ledger_api_pb2.LedgerApiMessage.Raw_Transaction_Performative()  # type: ignore
            raw_transaction = msg.raw_transaction
            RawTransaction.encode(performative.raw_transaction, raw_transaction)
            ledger_api_msg.raw_transaction.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.TRANSACTION_DIGEST:
            performative = ledger_api_pb2.LedgerApiMessage.Transaction_Digest_Performative()  # type: ignore
            transaction_digest = msg.transaction_digest
            TransactionDigest.encode(
                performative.transaction_digest, transaction_digest
            )
            ledger_api_msg.transaction_digest.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.TRANSACTION_RECEIPT:
            performative = ledger_api_pb2.LedgerApiMessage.Transaction_Receipt_Performative()  # type: ignore
            transaction_receipt = msg.transaction_receipt
            TransactionReceipt.encode(
                performative.transaction_receipt, transaction_receipt
            )
            ledger_api_msg.transaction_receipt.CopyFrom(performative)
        elif performative_id == LedgerApiMessage.Performative.ERROR:
            performative = ledger_api_pb2.LedgerApiMessage.Error_Performative()  # type: ignore
            code = msg.code
            performative.code = code
            if msg.is_set("message"):
                performative.message_is_set = True
                message = msg.message
                performative.message = message
            if msg.is_set("data"):
                performative.data_is_set = True
                data = msg.data
                performative.data = data
            ledger_api_msg.error.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        ledger_api_bytes = ledger_api_msg.SerializeToString()
        return ledger_api_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'LedgerApi' message.

        :param obj: the bytes object.
        :return: the 'LedgerApi' message.
        """
        ledger_api_pb = ledger_api_pb2.LedgerApiMessage()
        ledger_api_pb.ParseFromString(obj)
        message_id = ledger_api_pb.message_id
        dialogue_reference = (
            ledger_api_pb.dialogue_starter_reference,
            ledger_api_pb.dialogue_responder_reference,
        )
        target = ledger_api_pb.target

        performative = ledger_api_pb.WhichOneof("performative")
        performative_id = LedgerApiMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == LedgerApiMessage.Performative.GET_BALANCE:
            ledger_id = ledger_api_pb.get_balance.ledger_id
            performative_content["ledger_id"] = ledger_id
            address = ledger_api_pb.get_balance.address
            performative_content["address"] = address
        elif performative_id == LedgerApiMessage.Performative.GET_RAW_TRANSACTION:
            pb2_terms = ledger_api_pb.get_raw_transaction.terms
            terms = Terms.decode(pb2_terms)
            performative_content["terms"] = terms
        elif performative_id == LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION:
            pb2_signed_transaction = (
                ledger_api_pb.send_signed_transaction.signed_transaction
            )
            signed_transaction = SignedTransaction.decode(pb2_signed_transaction)
            performative_content["signed_transaction"] = signed_transaction
        elif performative_id == LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT:
            pb2_transaction_digest = (
                ledger_api_pb.get_transaction_receipt.transaction_digest
            )
            transaction_digest = TransactionDigest.decode(pb2_transaction_digest)
            performative_content["transaction_digest"] = transaction_digest
        elif performative_id == LedgerApiMessage.Performative.BALANCE:
            ledger_id = ledger_api_pb.balance.ledger_id
            performative_content["ledger_id"] = ledger_id
            balance = ledger_api_pb.balance.balance
            performative_content["balance"] = balance
        elif performative_id == LedgerApiMessage.Performative.RAW_TRANSACTION:
            pb2_raw_transaction = ledger_api_pb.raw_transaction.raw_transaction
            raw_transaction = RawTransaction.decode(pb2_raw_transaction)
            performative_content["raw_transaction"] = raw_transaction
        elif performative_id == LedgerApiMessage.Performative.TRANSACTION_DIGEST:
            pb2_transaction_digest = ledger_api_pb.transaction_digest.transaction_digest
            transaction_digest = TransactionDigest.decode(pb2_transaction_digest)
            performative_content["transaction_digest"] = transaction_digest
        elif performative_id == LedgerApiMessage.Performative.TRANSACTION_RECEIPT:
            pb2_transaction_receipt = (
                ledger_api_pb.transaction_receipt.transaction_receipt
            )
            transaction_receipt = TransactionReceipt.decode(pb2_transaction_receipt)
            performative_content["transaction_receipt"] = transaction_receipt
        elif performative_id == LedgerApiMessage.Performative.ERROR:
            code = ledger_api_pb.error.code
            performative_content["code"] = code
            if ledger_api_pb.error.message_is_set:
                message = ledger_api_pb.error.message
                performative_content["message"] = message
            if ledger_api_pb.error.data_is_set:
                data = ledger_api_pb.error.data
                performative_content["data"] = data
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return LedgerApiMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This contains the proxy agent class."""

import sys
from queue import Queue

import gym

from aea.agent import Agent
from aea.configurations.base import ConnectionConfig
from aea.helpers.base import locate
from aea.identity.base import Identity
from aea.mail.base import Envelope


sys.modules["packages.fetchai.connections.gym"] = locate(  # isort:skip
    "packages.fetchai.connections.gym"
)


from packages.fetchai.connections.gym.connection import (  # noqa: E402  # pylint: disable=wrong-import-position
    GymConnection,
)


ADDRESS = "some_address"


class ProxyAgent(Agent):
    """This class implements a proxy agent to be used by a proxy environment."""

    def __init__(self, name: str, gym_env: gym.Env, proxy_env_queue: Queue) -> None:
        """
        Instantiate the agent.

        :param name: the name of the agent
        :param gym_env: gym environment
        :param proxy_env_queue: the queue of the proxy environment
        :return: None
        """
        identity = Identity(name, ADDRESS)
        configuration = ConnectionConfig(connection_id=GymConnection.connection_id)
        super().__init__(
            identity,
            [GymConnection(gym_env, identity=identity, configuration=configuration)],
            period=0.01,
        )
        self.proxy_env_queue = proxy_env_queue

    def setup(self) -> None:
        """
        Set up the agent.

        :return: None
        """
        pass

    def act(self) -> None:
        """
        Perform actions.

        :return: None
        """
        pass

    def handle_envelope(self, envelope: Envelope) -> None:
        """
        Handle envelope.

        :param envelope: the envelope
        :return: None
        """
        if envelope is not None:
            self.proxy_env_queue.put(envelope)

    def teardown(self) -> None:
        """
        Tear down the agent.

        :return: None
        """
        pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for generator/validate.py module."""
import logging
from unittest import TestCase, mock

from aea.configurations.base import CRUDCollection, SpeechActContentConfig
from aea.protocols.generator.validate import (
    CONTENT_NAME_REGEX_PATTERN,
    END_STATE_REGEX_PATTERN,
    PERFORMATIVE_REGEX_PATTERN,
    ROLE_REGEX_PATTERN,
    _has_brackets,
    _is_reserved_name,
    _is_valid_content_type_format,
    _is_valid_ct,
    _is_valid_dict,
    _is_valid_list,
    _is_valid_optional,
    _is_valid_pt,
    _is_valid_regex,
    _is_valid_set,
    _is_valid_union,
    _validate_content_name,
    _validate_content_type,
    _validate_dialogue_section,
    _validate_end_states,
    _validate_initiation,
    _validate_performatives,
    _validate_protocol_buffer_schema_code_snippets,
    _validate_reply,
    _validate_roles,
    _validate_speech_acts_section,
    _validate_termination,
    validate,
)


logger = logging.getLogger("aea")
logging.basicConfig(level=logging.INFO)


class TestValidate(TestCase):
    """Test for generator/validate.py."""

    def test_is_reserved_name(self):
        """Test for the '_is_reserved_name' method."""
        invalid_content_name_1 = "body"
        assert _is_reserved_name(invalid_content_name_1) is True

        invalid_content_name_2 = "message_id"
        assert _is_reserved_name(invalid_content_name_2) is True

        invalid_content_name_3 = "dialogue_reference"
        assert _is_reserved_name(invalid_content_name_3) is True

        invalid_content_name_4 = "target"
        assert _is_reserved_name(invalid_content_name_4) is True

        invalid_content_name_5 = "performative"
        assert _is_reserved_name(invalid_content_name_5) is True

        valid_content_nam_1 = "content_name"
        assert _is_reserved_name(valid_content_nam_1) is False

        valid_content_name_2 = "query"
        assert _is_reserved_name(valid_content_name_2) is False

        valid_content_name_3 = "ThiSiSAConTEnT234"
        assert _is_reserved_name(valid_content_name_3) is False

    def test_is_valid_regex(self):
        """Test for the '_is_valid_regex' method."""
        regex_1 = "^[0-9][a-zA-Z0-9]*[A-Z]$"

        valid_text_1 = "53453hKb35nDkG"
        assert _is_valid_regex(regex_1, valid_text_1) is True

        invalid_text_1 = "hKbnDkG"
        assert _is_valid_regex(regex_1, invalid_text_1) is False

        invalid_text_2 = "4f nkG"
        assert _is_valid_regex(regex_1, invalid_text_2) is False

    def test_has_brackets(self):
        """Test for the '_has_brackets' method."""
        valid_content_type_1 = "pt:set[pt:int]"
        assert _has_brackets(valid_content_type_1) is True

        valid_content_type_2 = "pt:union[hskdjf-8768&^${]hsdkjhfk]"
        assert _has_brackets(valid_content_type_2) is True

        valid_content_type_3 = "pt:optional[[]]"
        assert _has_brackets(valid_content_type_3) is True

        ###################################################

        invalid_content_type_1 = "ct:set[pt:int]"
        with self.assertRaises(SyntaxError) as cm:
            _has_brackets(invalid_content_type_1)
        self.assertEqual(
            str(cm.exception), "Content type must be a compositional type!"
        )

        invalid_content_type_2 = "pt:tuple[pt:float]"
        with self.assertRaises(SyntaxError) as cm:
            _has_brackets(invalid_content_type_2)
        self.assertEqual(
            str(cm.exception), "Content type must be a compositional type!"
        )

        invalid_content_type_3 = "pt:optinal[pt:bool]"
        with self.assertRaises(SyntaxError) as cm:
            _has_brackets(invalid_content_type_3)
        self.assertEqual(
            str(cm.exception), "Content type must be a compositional type!"
        )

        ###################################################

        invalid_content_type_4 = "pt:optional{}"
        assert _has_brackets(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:set[]7657"
        assert _has_brackets(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:union [pt:int, pt:bool]"
        assert _has_brackets(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:dict[pt:int, pt:bool] "
        assert _has_brackets(invalid_content_type_7) is False

    def test_is_valid_ct(self):
        """Test for the '_is_valid_ct' method."""
        valid_content_type_1 = "ct:DataModel"
        assert _is_valid_ct(valid_content_type_1) is True

        valid_content_type_2 = "ct:ThisIsACustomContent"
        assert _is_valid_ct(valid_content_type_2) is True

        valid_content_type_3 = "ct:Query"
        assert _is_valid_ct(valid_content_type_3) is True

        valid_content_type_4 = "   ct:Proposal "
        assert _is_valid_ct(valid_content_type_4) is True

        valid_content_type_5 = "ct:DSA"
        assert _is_valid_ct(valid_content_type_5) is True

        valid_content_type_6 = "ct:DataF"
        assert _is_valid_ct(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "ct:data"
        assert _is_valid_ct(invalid_content_type_1) is False

        invalid_content_type_2 = "Model"
        assert _is_valid_ct(invalid_content_type_2) is False

        invalid_content_type_3 = "ct: DataModel"
        assert _is_valid_ct(invalid_content_type_3) is False

        invalid_content_type_4 = "ct:E3"
        assert _is_valid_ct(invalid_content_type_4) is False

    def test_is_valid_pt(self):
        """Test for the '_is_valid_pt' method."""
        valid_content_type_1 = "pt:bytes"
        assert _is_valid_pt(valid_content_type_1) is True

        valid_content_type_2 = "pt:int"
        assert _is_valid_pt(valid_content_type_2) is True

        valid_content_type_3 = "pt:float"
        assert _is_valid_pt(valid_content_type_3) is True

        valid_content_type_4 = "pt:bool"
        assert _is_valid_pt(valid_content_type_4) is True

        valid_content_type_5 = "pt:str"
        assert _is_valid_pt(valid_content_type_5) is True

        valid_content_type_6 = "  pt:int  "
        assert _is_valid_pt(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "pt:integer"
        assert _is_valid_pt(invalid_content_type_1) is False

        invalid_content_type_2 = "bool"
        assert _is_valid_pt(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: str"
        assert _is_valid_pt(invalid_content_type_3) is False

        invalid_content_type_4 = "pt;float"
        assert _is_valid_pt(invalid_content_type_4) is False

    def test_is_valid_set(self):
        """Test for the '_is_valid_set' method."""
        valid_content_type_1 = "pt:set[pt:bytes]"
        assert _is_valid_set(valid_content_type_1) is True

        valid_content_type_2 = "pt:set[pt:int]"
        assert _is_valid_set(valid_content_type_2) is True

        valid_content_type_3 = "pt:set[pt:float]"
        assert _is_valid_set(valid_content_type_3) is True

        valid_content_type_4 = "pt:set[pt:bool]"
        assert _is_valid_set(valid_content_type_4) is True

        valid_content_type_5 = "pt:set[pt:str]"
        assert _is_valid_set(valid_content_type_5) is True

        valid_content_type_6 = " pt:set[   pt:int ]   "
        assert _is_valid_set(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "pt:frozenset[pt:int]"
        assert _is_valid_set(invalid_content_type_1) is False

        invalid_content_type_2 = "set[pt:int]"
        assert _is_valid_set(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: set[pt:int]"
        assert _is_valid_set(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:set[integer]"
        assert _is_valid_set(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:set[int]"
        assert _is_valid_set(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:set{int]"
        assert _is_valid_set(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:set[pt:int, pt:str]"
        assert _is_valid_set(invalid_content_type_7) is False

        invalid_content_type_8 = "pt:set[]"
        assert _is_valid_set(invalid_content_type_8) is False

        invalid_content_type_9 = "pt:set[pt:list[pt:int, pt:list[pt:bool]]"
        assert _is_valid_set(invalid_content_type_9) is False

        invalid_content_type_10 = "pt:set"
        assert _is_valid_set(invalid_content_type_10) is False

    def test_is_valid_list(self):
        """Test for the '_is_valid_list' method."""
        valid_content_type_1 = "pt:list[pt:bytes]"
        assert _is_valid_list(valid_content_type_1) is True

        valid_content_type_2 = "pt:list[pt:int]"
        assert _is_valid_list(valid_content_type_2) is True

        valid_content_type_3 = "pt:list[pt:float]"
        assert _is_valid_list(valid_content_type_3) is True

        valid_content_type_4 = "pt:list[pt:bool]"
        assert _is_valid_list(valid_content_type_4) is True

        valid_content_type_5 = "pt:list[pt:str]"
        assert _is_valid_list(valid_content_type_5) is True

        valid_content_type_6 = " pt:list[   pt:bool ]   "
        assert _is_valid_list(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "pt:tuple[pt:bytes]"
        assert _is_valid_list(invalid_content_type_1) is False

        invalid_content_type_2 = "list[pt:bool]"
        assert _is_valid_list(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: list[pt:float]"
        assert _is_valid_list(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:list[string]"
        assert _is_valid_list(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:list[bool]"
        assert _is_valid_list(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:list[bytes"
        assert _is_valid_list(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:list[pt:float, pt:bool]"
        assert _is_valid_list(invalid_content_type_7) is False

        invalid_content_type_8 = "pt:list[]"
        assert _is_valid_list(invalid_content_type_8) is False

        invalid_content_type_9 = "pt:list[pt:set[pt:bool, pt:set[pt:str]]"
        assert _is_valid_list(invalid_content_type_9) is False

        invalid_content_type_10 = "pt:list"
        assert _is_valid_list(invalid_content_type_10) is False

    def test_is_valid_dict(self):
        """Test for the '_is_valid_dict' method."""
        valid_content_type_1 = "pt:dict[pt:bytes, pt:int]"
        assert _is_valid_dict(valid_content_type_1) is True

        valid_content_type_2 = "pt:dict[pt:int, pt:int]"
        assert _is_valid_dict(valid_content_type_2) is True

        valid_content_type_3 = "pt:dict[pt:float, pt:str]"
        assert _is_valid_dict(valid_content_type_3) is True

        valid_content_type_4 = "pt:dict[pt:bool, pt:str]"
        assert _is_valid_dict(valid_content_type_4) is True

        valid_content_type_5 = "pt:dict[pt:bool,pt:float]"
        assert _is_valid_dict(valid_content_type_5) is True

        valid_content_type_6 = "   pt:dict[  pt:bytes  ,   pt:int   ] "
        assert _is_valid_dict(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "pt:map[pt:bool, pt:str]"
        assert _is_valid_dict(invalid_content_type_1) is False

        invalid_content_type_2 = "dict[pt:int, pt:float]"
        assert _is_valid_dict(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: dict[pt:bytes, pt:bool]"
        assert _is_valid_dict(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:dict[float, pt:str]"
        assert _is_valid_dict(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:dict[pt:bool, pt:integer]"
        assert _is_valid_dict(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:dict(pt:boolean, pt:int"
        assert _is_valid_dict(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:dict[pt:boolean]"
        assert _is_valid_dict(invalid_content_type_7) is False

        invalid_content_type_8 = "pt:dict[]"
        assert _is_valid_dict(invalid_content_type_8) is False

        invalid_content_type_9 = "pt:dict[pt:str, pt:float, pt:int, pt:bytes]"
        assert _is_valid_dict(invalid_content_type_9) is False

        invalid_content_type_10 = "pt:dict[pt:set[pt:bool, pt:str]"
        assert _is_valid_dict(invalid_content_type_10) is False

        invalid_content_type_11 = "pt:dict"
        assert _is_valid_dict(invalid_content_type_11) is False

    def test_is_valid_union(self):
        """Test for the '_is_valid_union' method."""
        valid_content_type_1 = (
            "pt:union[pt:bytes, pt:int, pt:float, pt:bool, pt:str, pt:set[pt:bytes], "
            "pt:set[pt:int], pt:set[pt:float], pt:set[pt:bool], pt:set[pt:str], "
            "pt:list[pt:bytes], pt:list[pt:int], pt:list[pt:float], pt:list[pt:bool], pt:list[pt:str], "
            "pt:dict[pt:bytes, pt:bytes],    pt:dict[  pt:bytes  ,   pt:int   ]  , pt:dict[pt:bytes, pt:float], pt:dict[pt:bytes, pt:bool], pt:dict[pt:bytes, pt:str], "
            "pt:dict[pt:int, pt:bytes], pt:dict[pt:int, pt:int], pt:dict[pt:int, pt:float], pt:dict[pt:int, pt:bool], pt:dict[pt:int, pt:str], "
            "pt:dict[pt:float, pt:bytes], pt:dict[pt:float, pt:int], pt:dict[pt:float, pt:float], pt:dict[pt:float, pt:bool], pt:dict[pt:float, pt:str], "
            "pt:dict[pt:bool, pt:bytes], pt:dict[pt:bool, pt:int], pt:dict[pt:bool,pt:float], pt:dict[pt:bool, pt:bool], pt:dict[pt:bool, pt:str], "
            "pt:dict[pt:str, pt:bytes], pt:dict[pt:str, pt:int], pt:dict[pt:str,pt:float], pt:dict[pt:str, pt:bool], pt:dict[pt:str, pt:str]]"
        )
        assert _is_valid_union(valid_content_type_1) is True

        valid_content_type_2 = "pt:union[pt:bytes, pt:set[pt:int]]"
        assert _is_valid_union(valid_content_type_2) is True

        valid_content_type_3 = "pt:union[pt:float, pt:bool]"
        assert _is_valid_union(valid_content_type_3) is True

        valid_content_type_4 = "pt:union[pt:set[pt:int], pt:set[pt:float]]"
        assert _is_valid_union(valid_content_type_4) is True

        valid_content_type_5 = "pt:union[pt:bool,pt:bytes]"
        assert _is_valid_union(valid_content_type_5) is True

        valid_content_type_6 = "   pt:union[  pt:bytes  ,   pt:set[  pt:int  ]   ] "
        assert _is_valid_union(valid_content_type_6) is True

        ###################################################

        invalid_content_type_1 = "pt:onion[pt:bool, pt:str]"
        assert _is_valid_union(invalid_content_type_1) is False

        invalid_content_type_2 = "union[pt:int, pt:float]"
        assert _is_valid_union(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: union[pt:set[pt:int], pt:bool]"
        assert _is_valid_union(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:union[float, pt:str"
        assert _is_valid_union(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:union[pt:int, pt:dict[pt:str, pt:bool]"
        assert _is_valid_union(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:union{pt:boolean, pt:int]"
        assert _is_valid_union(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:union[pt:boolean]"
        assert _is_valid_union(invalid_content_type_7) is False

        invalid_content_type_8 = "pt:union[]"
        assert _is_valid_union(invalid_content_type_8) is False

        invalid_content_type_9 = "pt:union[pt:str, pt:int, pt:str]"
        assert _is_valid_union(invalid_content_type_9) is False

        invalid_content_type_10 = "pt:union[pt:set[pt:integer], pt:float]"
        assert _is_valid_union(invalid_content_type_10) is False

        invalid_content_type_11 = (
            "pt:union[pt:dict[pt:set[pt:bool]], pt:list[pt:set[pt:str]]]"
        )
        assert _is_valid_union(invalid_content_type_11) is False

        invalid_content_type_12 = "pt:union"
        assert _is_valid_union(invalid_content_type_12) is False

    def test_is_valid_optional(self):
        """Test for the '_is_valid_optional' method."""
        valid_content_type_1 = (
            "pt:optional[pt:union[pt:bytes, pt:int, pt:float, pt:bool, pt:str, pt:set[pt:bytes], "
            "pt:set[pt:int], pt:set[pt:float], pt:set[pt:bool], pt:set[pt:str], "
            "pt:list[pt:bytes], pt:list[pt:int], pt:list[pt:float], pt:list[pt:bool], pt:list[pt:str], "
            "pt:dict[pt:bytes, pt:bytes],    pt:dict[  pt:bytes  ,   pt:int   ]  , pt:dict[pt:bytes, pt:float], pt:dict[pt:bytes, pt:bool], pt:dict[pt:bytes, pt:str], "
            "pt:dict[pt:int, pt:bytes], pt:dict[pt:int, pt:int], pt:dict[pt:int, pt:float], pt:dict[pt:int, pt:bool], pt:dict[pt:int, pt:str], "
            "pt:dict[pt:float, pt:bytes], pt:dict[pt:float, pt:int], pt:dict[pt:float, pt:float], pt:dict[pt:float, pt:bool], pt:dict[pt:float, pt:str], "
            "pt:dict[pt:bool, pt:bytes], pt:dict[pt:bool, pt:int], pt:dict[pt:bool,pt:float], pt:dict[pt:bool, pt:bool], pt:dict[pt:bool, pt:str], "
            "pt:dict[pt:str, pt:bytes], pt:dict[pt:str, pt:int], pt:dict[pt:str,pt:float], pt:dict[pt:str, pt:bool], pt:dict[pt:str, pt:str]]]"
        )
        assert _is_valid_optional(valid_content_type_1) is True

        valid_content_type_2 = "pt:optional[pt:union[pt:bytes, pt:set[pt:int]]]"
        assert _is_valid_optional(valid_content_type_2) is True

        valid_content_type_3 = "pt:optional[pt:bytes]"
        assert _is_valid_optional(valid_content_type_3) is True

        valid_content_type_4 = "pt:optional[pt:int]"
        assert _is_valid_optional(valid_content_type_4) is True

        valid_content_type_5 = "pt:optional[pt:float]"
        assert _is_valid_optional(valid_content_type_5) is True

        valid_content_type_6 = "pt:optional[pt:bool]"
        assert _is_valid_optional(valid_content_type_6) is True

        valid_content_type_7 = "pt:optional[pt:str]"
        assert _is_valid_optional(valid_content_type_7) is True

        valid_content_type_8 = "pt:optional[pt:set[pt:bytes]]"
        assert _is_valid_optional(valid_content_type_8) is True

        valid_content_type_9 = "pt:optional[pt:list[pt:int]]"
        assert _is_valid_optional(valid_content_type_9) is True

        valid_content_type_10 = (
            "     pt:optional[  pt:dict[   pt:float   ,  pt:bool ]   ] "
        )
        assert _is_valid_optional(valid_content_type_10) is True

        ###################################################

        invalid_content_type_1 = "pt:optinal[pt:bytes]"
        assert _is_valid_optional(invalid_content_type_1) is False

        invalid_content_type_2 = "optional[pt:int]"
        assert _is_valid_optional(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: optional[pt:float]"
        assert _is_valid_optional(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:optional[bool]"
        assert _is_valid_optional(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:optional[pt:str"
        assert _is_valid_optional(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:optional{pt:set[pt:int]]"
        assert _is_valid_optional(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:optional[pt:string]"
        assert _is_valid_optional(invalid_content_type_7) is False

        invalid_content_type_8 = "pt:optional[]"
        assert _is_valid_optional(invalid_content_type_8) is False

        invalid_content_type_9 = "pt:optional[pt:str, pt:int, pt:list[pt:bool]]"
        assert _is_valid_optional(invalid_content_type_9) is False

        invalid_content_type_10 = "pt:optional[pt:list[pt:boolean]]"
        assert _is_valid_optional(invalid_content_type_10) is False

        invalid_content_type_11 = "pt:optional[pt:dict[pt:set[pt:int]]]"
        assert _is_valid_optional(invalid_content_type_11) is False

        invalid_content_type_12 = "pt:optional"
        assert _is_valid_optional(invalid_content_type_12) is False

    def test_is_valid_content_type_format(self):
        """Test for the '_is_valid_content_type_format' method."""
        valid_content_type_1 = "ct:DataModel"
        assert _is_valid_content_type_format(valid_content_type_1) is True

        valid_content_type_2 = "pt:int"
        assert _is_valid_content_type_format(valid_content_type_2) is True

        valid_content_type_3 = "pt:set[pt:float]"
        assert _is_valid_content_type_format(valid_content_type_3) is True

        valid_content_type_4 = "pt:list[pt:bool]"
        assert _is_valid_content_type_format(valid_content_type_4) is True

        valid_content_type_5 = "pt:dict[pt:bool,pt:float]"
        assert _is_valid_content_type_format(valid_content_type_5) is True

        valid_content_type_6 = (
            "pt:optional[pt:union[pt:bytes, pt:int, pt:float, pt:bool, pt:str, pt:set[pt:bytes], "
            "pt:set[pt:int], pt:set[pt:float], pt:set[pt:bool], pt:set[pt:str], "
            "pt:list[pt:bytes], pt:list[pt:int], pt:list[pt:float], pt:list[pt:bool], pt:list[pt:str], "
            "pt:dict[pt:bytes, pt:bytes],    pt:dict[  pt:bytes  ,   pt:int   ]  , pt:dict[pt:bytes, pt:float], pt:dict[pt:bytes, pt:bool], pt:dict[pt:bytes, pt:str], "
            "pt:dict[pt:int, pt:bytes], pt:dict[pt:int, pt:int], pt:dict[pt:int, pt:float], pt:dict[pt:int, pt:bool], pt:dict[pt:int, pt:str], "
            "pt:dict[pt:float, pt:bytes], pt:dict[pt:float, pt:int], pt:dict[pt:float, pt:float], pt:dict[pt:float, pt:bool], pt:dict[pt:float, pt:str], "
            "pt:dict[pt:bool, pt:bytes], pt:dict[pt:bool, pt:int], pt:dict[pt:bool,pt:float], pt:dict[pt:bool, pt:bool], pt:dict[pt:bool, pt:str], "
            "pt:dict[pt:str, pt:bytes], pt:dict[pt:str, pt:int], pt:dict[pt:str,pt:float], pt:dict[pt:str, pt:bool], pt:dict[pt:str, pt:str]]]"
        )
        assert _is_valid_content_type_format(valid_content_type_6) is True

        valid_content_type_7 = (
            "     pt:optional[  pt:dict[   pt:float   ,  pt:bool ]   ] "
        )
        assert _is_valid_content_type_format(valid_content_type_7) is True

        ###################################################

        invalid_content_type_1 = "ct:data"
        assert _is_valid_content_type_format(invalid_content_type_1) is False

        invalid_content_type_2 = "bool"
        assert _is_valid_content_type_format(invalid_content_type_2) is False

        invalid_content_type_3 = "pt: set[pt:int]"
        assert _is_valid_content_type_format(invalid_content_type_3) is False

        invalid_content_type_4 = "pt:list[string]"
        assert _is_valid_content_type_format(invalid_content_type_4) is False

        invalid_content_type_5 = "pt:dict[pt:bool, pt:integer]"
        assert _is_valid_content_type_format(invalid_content_type_5) is False

        invalid_content_type_6 = "pt:union{pt:boolean, pt:int]"
        assert _is_valid_content_type_format(invalid_content_type_6) is False

        invalid_content_type_7 = "pt:optional[pt:str, pt:int, pt:list[pt:bool]]"
        assert _is_valid_content_type_format(invalid_content_type_7) is False

    def test_validate_performatives(self):
        """Test for the '_validate_performatives' method."""
        valid_content_type_1 = "offer"
        valid_result_1, valid_msg_1 = _validate_performatives(valid_content_type_1)
        assert valid_result_1 is True
        assert valid_msg_1 == "Performative '{}' is valid.".format(valid_content_type_1)

        valid_content_type_2 = "send_HTTP_message"
        valid_result_2, valid_msg_2 = _validate_performatives(valid_content_type_2)
        assert valid_result_2 is True
        assert valid_msg_2 == "Performative '{}' is valid.".format(valid_content_type_2)

        valid_content_type_3 = "request_2PL"
        valid_result_3, valid_msg_3 = _validate_performatives(valid_content_type_3)
        assert valid_result_3 is True
        assert valid_msg_3 == "Performative '{}' is valid.".format(valid_content_type_3)

        valid_content_type_4 = "argue"
        valid_result_4, valid_msg_4 = _validate_performatives(valid_content_type_4)
        assert valid_result_4 is True
        assert valid_msg_4 == "Performative '{}' is valid.".format(valid_content_type_4)

        ###################################################

        invalid_content_type_1 = "_offer"
        invalid_result_1, invalid_msg_1 = _validate_performatives(
            invalid_content_type_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_content_type_1, PERFORMATIVE_REGEX_PATTERN
            )
        )

        invalid_content_type_2 = "request_"
        invalid_result_2, invalid_msg_2 = _validate_performatives(
            invalid_content_type_2
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_content_type_2, PERFORMATIVE_REGEX_PATTERN
            )
        )

        invalid_content_type_3 = "_query_"
        invalid_result_3, invalid_msg_3 = _validate_performatives(
            invalid_content_type_3
        )
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_content_type_3, PERFORMATIVE_REGEX_PATTERN
            )
        )

        invalid_content_type_4 = "$end"
        invalid_result_4, invalid_msg_4 = _validate_performatives(
            invalid_content_type_4
        )
        assert invalid_result_4 is False
        assert (
            invalid_msg_4
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_content_type_4, PERFORMATIVE_REGEX_PATTERN
            )
        )

        invalid_content_type_5 = "create()"
        invalid_result_5, invalid_msg_5 = _validate_performatives(
            invalid_content_type_5
        )
        assert invalid_result_5 is False
        assert (
            invalid_msg_5
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_content_type_5, PERFORMATIVE_REGEX_PATTERN
            )
        )

        invalid_content_type_6 = "body"
        invalid_result_6, invalid_msg_6 = _validate_performatives(
            invalid_content_type_6
        )
        assert invalid_result_6 is False
        assert (
            invalid_msg_6
            == "Invalid name for performative '{}'. This name is reserved.".format(
                invalid_content_type_6,
            )
        )

        invalid_content_type_7 = "message_id"
        invalid_result_7, invalid_msg_7 = _validate_performatives(
            invalid_content_type_7
        )
        assert invalid_result_7 is False
        assert (
            invalid_msg_6
            == "Invalid name for performative '{}'. This name is reserved.".format(
                invalid_content_type_6,
            )
        )

    def test_validate_content_name(self):
        """Test for the '_validate_content_name' method."""
        performative = "some_performative"

        valid_content_type_1 = "content"
        valid_result_1, valid_msg_1 = _validate_content_name(
            valid_content_type_1, performative
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Content name '{}' of performative '{}' is valid.".format(
            valid_content_type_1, performative
        )

        valid_content_type_2 = "HTTP_msg_name"
        valid_result_2, valid_msg_2 = _validate_content_name(
            valid_content_type_2, performative
        )
        assert valid_result_2 is True
        assert valid_msg_2 == "Content name '{}' of performative '{}' is valid.".format(
            valid_content_type_2, performative
        )

        valid_content_type_3 = "number_of_3PLs"
        valid_result_3, valid_msg_3 = _validate_content_name(
            valid_content_type_3, performative
        )
        assert valid_result_3 is True
        assert valid_msg_3 == "Content name '{}' of performative '{}' is valid.".format(
            valid_content_type_3, performative
        )

        valid_content_type_4 = "model"
        valid_result_4, valid_msg_4 = _validate_content_name(
            valid_content_type_4, performative
        )
        assert valid_result_4 is True
        assert valid_msg_4 == "Content name '{}' of performative '{}' is valid.".format(
            valid_content_type_4, performative
        )

        ###################################################

        invalid_content_type_1 = "_content"
        invalid_result_1, invalid_msg_1 = _validate_content_name(
            invalid_content_type_1, performative
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                invalid_content_type_1, performative, CONTENT_NAME_REGEX_PATTERN
            )
        )

        invalid_content_type_2 = "content_"
        invalid_result_2, invalid_msg_2 = _validate_content_name(
            invalid_content_type_2, performative
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                invalid_content_type_2, performative, CONTENT_NAME_REGEX_PATTERN
            )
        )

        invalid_content_type_3 = "_content_"
        invalid_result_3, invalid_msg_3 = _validate_content_name(
            invalid_content_type_3, performative
        )
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                invalid_content_type_3, performative, CONTENT_NAME_REGEX_PATTERN
            )
        )

        invalid_content_type_4 = "con^en^"
        invalid_result_4, invalid_msg_4 = _validate_content_name(
            invalid_content_type_4, performative
        )
        assert invalid_result_4 is False
        assert (
            invalid_msg_4
            == "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                invalid_content_type_4, performative, CONTENT_NAME_REGEX_PATTERN
            )
        )

        invalid_content_type_5 = "some_content()"
        invalid_result_5, invalid_msg_5 = _validate_content_name(
            invalid_content_type_5, performative
        )
        assert invalid_result_5 is False
        assert (
            invalid_msg_5
            == "Invalid name for content '{}' of performative '{}'. Content names must match the following regular expression: {} ".format(
                invalid_content_type_5, performative, CONTENT_NAME_REGEX_PATTERN
            )
        )

        invalid_content_type_6 = "target"
        invalid_result_6, invalid_msg_6 = _validate_content_name(
            invalid_content_type_6, performative
        )
        assert invalid_result_6 is False
        assert (
            invalid_msg_6
            == "Invalid name for content '{}' of performative '{}'. This name is reserved.".format(
                invalid_content_type_6, performative,
            )
        )

        invalid_content_type_7 = "performative"
        invalid_result_7, invalid_msg_7 = _validate_content_name(
            invalid_content_type_7, performative
        )
        assert invalid_result_7 is False
        assert (
            invalid_msg_7
            == "Invalid name for content '{}' of performative '{}'. This name is reserved.".format(
                invalid_content_type_7, performative,
            )
        )

    def test_validate_content_type(self):
        """Test for the '_validate_content_type' method."""
        performative = "some_performative"
        content_name = "some_content_name"

        valid_content_type_1 = "ct:DataModel"
        valid_result_1, valid_msg_1 = _validate_content_type(
            valid_content_type_1, content_name, performative
        )
        assert valid_result_1 is True
        assert (
            valid_msg_1
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_2 = "pt:int"
        valid_result_2, valid_msg_2 = _validate_content_type(
            valid_content_type_2, content_name, performative
        )
        assert valid_result_2 is True
        assert (
            valid_msg_2
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_3 = "pt:set[pt:float]"
        valid_result_3, valid_msg_3 = _validate_content_type(
            valid_content_type_3, content_name, performative
        )
        assert valid_result_3 is True
        assert (
            valid_msg_3
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_4 = "pt:list[pt:bool]"
        valid_result_4, valid_msg_4 = _validate_content_type(
            valid_content_type_4, content_name, performative
        )
        assert valid_result_4 is True
        assert (
            valid_msg_4
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_5 = "pt:dict[pt:bool,pt:float]"
        valid_result_5, valid_msg_5 = _validate_content_type(
            valid_content_type_5, content_name, performative
        )
        assert valid_result_5 is True
        assert (
            valid_msg_5
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_6 = (
            "pt:optional[pt:union[pt:bytes, pt:int, pt:float, pt:bool, pt:str, pt:set[pt:bytes], "
            "pt:set[pt:int], pt:set[pt:float], pt:set[pt:bool], pt:set[pt:str], "
            "pt:list[pt:bytes], pt:list[pt:int], pt:list[pt:float], pt:list[pt:bool], pt:list[pt:str], "
            "pt:dict[pt:bytes, pt:bytes],    pt:dict[  pt:bytes  ,   pt:int   ]  , pt:dict[pt:bytes, pt:float], pt:dict[pt:bytes, pt:bool], pt:dict[pt:bytes, pt:str], "
            "pt:dict[pt:int, pt:bytes], pt:dict[pt:int, pt:int], pt:dict[pt:int, pt:float], pt:dict[pt:int, pt:bool], pt:dict[pt:int, pt:str], "
            "pt:dict[pt:float, pt:bytes], pt:dict[pt:float, pt:int], pt:dict[pt:float, pt:float], pt:dict[pt:float, pt:bool], pt:dict[pt:float, pt:str], "
            "pt:dict[pt:bool, pt:bytes], pt:dict[pt:bool, pt:int], pt:dict[pt:bool,pt:float], pt:dict[pt:bool, pt:bool], pt:dict[pt:bool, pt:str], "
            "pt:dict[pt:str, pt:bytes], pt:dict[pt:str, pt:int], pt:dict[pt:str,pt:float], pt:dict[pt:str, pt:bool], pt:dict[pt:str, pt:str]]]"
        )
        valid_result_6, valid_msg_6 = _validate_content_type(
            valid_content_type_6, content_name, performative
        )
        assert valid_result_6 is True
        assert (
            valid_msg_6
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        valid_content_type_7 = (
            "     pt:optional[  pt:dict[   pt:float   ,  pt:bool ]   ] "
        )
        valid_result_7, valid_msg_7 = _validate_content_type(
            valid_content_type_7, content_name, performative
        )
        assert valid_result_7 is True
        assert (
            valid_msg_7
            == "Type of content '{}' of performative '{}' is valid.".format(
                content_name, performative
            )
        )

        ###################################################

        invalid_content_type_1 = "ct:data"
        invalid_result_1, invalid_msg_1 = _validate_content_type(
            invalid_content_type_1, content_name, performative
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_2 = "bool"
        invalid_result_2, invalid_msg_2 = _validate_content_type(
            invalid_content_type_2, content_name, performative
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_3 = "pt: set[pt:int]"
        invalid_result_3, invalid_msg_3 = _validate_content_type(
            invalid_content_type_3, content_name, performative
        )
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_4 = "pt:list[string]"
        invalid_result_4, invalid_msg_4 = _validate_content_type(
            invalid_content_type_4, content_name, performative
        )
        assert invalid_result_4 is False
        assert (
            invalid_msg_4
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_5 = "pt:dict[pt:bool, pt:integer]"
        invalid_result_5, invalid_msg_5 = _validate_content_type(
            invalid_content_type_5, content_name, performative
        )
        assert invalid_result_5 is False
        assert (
            invalid_msg_5
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_6 = "pt:union{pt:boolean, pt:int]"
        invalid_result_6, invalid_msg_6 = _validate_content_type(
            invalid_content_type_6, content_name, performative
        )
        assert invalid_result_6 is False
        assert (
            invalid_msg_6
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

        invalid_content_type_7 = "pt:optional[pt:str, pt:int, pt:list[pt:bool]]"
        invalid_result_7, invalid_msg_7 = _validate_content_type(
            invalid_content_type_7, content_name, performative
        )
        assert invalid_result_7 is False
        assert (
            invalid_msg_7
            == "Invalid type for content '{}' of performative '{}'. See documentation for the correct format of specification types.".format(
                content_name, performative,
            )
        )

    @mock.patch("aea.configurations.base.ProtocolSpecification",)
    def test_validate_speech_acts_section(self, mocked_spec):
        """Test for the '_validate_speech_acts_section' method."""
        valid_speech_act_content_config_1 = SpeechActContentConfig(
            content_1="ct:CustomType", content_2="pt:int"
        )
        valid_speech_act_content_config_2 = SpeechActContentConfig(
            content_3="ct:DataModel"
        )
        valid_speech_act_content_config_3 = SpeechActContentConfig()

        speech_act_1 = CRUDCollection()
        speech_act_1.create("perm_1", valid_speech_act_content_config_1)
        speech_act_1.create("perm_2", valid_speech_act_content_config_2)
        speech_act_1.create("perm_3", valid_speech_act_content_config_3)

        mocked_spec.speech_acts = speech_act_1

        (
            valid_result_1,
            valid_msg_1,
            valid_all_per_1,
            valid_all_content_1,
        ) = _validate_speech_acts_section(mocked_spec)
        assert valid_result_1 is True
        assert valid_msg_1 == "Speech-acts are valid."
        assert valid_all_per_1 == {"perm_1", "perm_2", "perm_3"}
        assert valid_all_content_1 == {"ct:CustomType", "ct:DataModel"}

        ###################################################

        speech_act_3 = CRUDCollection()
        invalid_perm = "_query_"
        speech_act_3.create(invalid_perm, valid_speech_act_content_config_1)

        mocked_spec.speech_acts = speech_act_3

        (
            invalid_result_1,
            invalid_msg_1,
            invalid_all_per_1,
            invalid_all_content_1,
        ) = _validate_speech_acts_section(mocked_spec)
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Invalid name for performative '{}'. Performative names must match the following regular expression: {} ".format(
                invalid_perm, PERFORMATIVE_REGEX_PATTERN
            )
        )
        assert invalid_all_per_1 is None
        assert invalid_all_content_1 is None

        invalid_speech_act_content_config_1 = SpeechActContentConfig(target="pt:int")
        speech_act_4 = CRUDCollection()
        valid_perm = "perm_1"
        speech_act_4.create(valid_perm, invalid_speech_act_content_config_1)

        mocked_spec.speech_acts = speech_act_4

        (
            invalid_result_2,
            invalid_msg_2,
            invalid_all_per_2,
            invalid_all_content_2,
        ) = _validate_speech_acts_section(mocked_spec)
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Invalid name for content '{}' of performative '{}'. This name is reserved.".format(
                "target", valid_perm,
            )
        )
        assert invalid_all_per_2 is None
        assert invalid_all_content_2 is None

        invalid_speech_act_content_config_2 = SpeechActContentConfig(
            conten_name_1="pt: set[pt:int]"
        )
        speech_act_5 = CRUDCollection()
        valid_perm = "perm_1"
        speech_act_5.create(valid_perm, invalid_speech_act_content_config_2)

        mocked_spec.speech_acts = speech_act_5

        (
            invalid_result_3,
            invalid_msg_3,
            invalid_all_per_3,
            invalid_all_content_3,
        ) = _validate_speech_acts_section(mocked_spec)
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Invalid type for content 'conten_name_1' of performative '{}'. See documentation for the correct format of specification types.".format(
                valid_perm,
            )
        )
        assert invalid_all_per_3 is None
        assert invalid_all_content_3 is None

        speech_act_6 = CRUDCollection()
        mocked_spec.speech_acts = speech_act_6

        (
            invalid_result_4,
            invalid_msg_4,
            invalid_all_per_4,
            invalid_all_content_4,
        ) = _validate_speech_acts_section(mocked_spec)
        assert invalid_result_4 is False
        assert invalid_msg_4 == "Speech-acts cannot be empty!"
        assert invalid_all_per_4 is None
        assert invalid_all_content_4 is None

    @mock.patch("aea.configurations.base.ProtocolSpecification",)
    def test_validate_protocol_buffer_schema_code_snippets(self, mocked_spec):
        """Test for the '_validate_protocol_buffer_schema_code_snippets' method."""
        valid_protobuf_snippet_1 = {
            "ct:DataModel": "bytes bytes_field = 1;\nint32 int_field = 2;\nfloat float_field = 3;\nbool bool_field = 4;\nstring str_field = 5;\nrepeated int32 set_field = 6;\nrepeated string list_field = 7;\nmap<int32, bool> dict_field = 8;\n"
        }
        valid_all_content_1 = {"ct:DataModel"}
        mocked_spec.protobuf_snippets = valid_protobuf_snippet_1

        valid_result_1, valid_msg_1, = _validate_protocol_buffer_schema_code_snippets(
            mocked_spec, valid_all_content_1
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Protobuf code snippet section is valid."

        valid_protobuf_snippet_2 = {}
        valid_all_content_2 = set()
        mocked_spec.protobuf_snippets = valid_protobuf_snippet_2

        valid_result_2, valid_msg_2, = _validate_protocol_buffer_schema_code_snippets(
            mocked_spec, valid_all_content_2
        )
        assert valid_result_2 is True
        assert valid_msg_2 == "Protobuf code snippet section is valid."

        ###################################################

        invalid_protobuf_snippet_1 = {
            "ct:DataModel": "bytes bytes_field = 1;\nint32 int_field = 2;\nfloat float_field = 3;\nbool bool_field = 4;\nstring str_field = 5;",
            "ct:Query": "bytes bytes_field = 1;",
        }
        invalid_all_content_1 = {"ct:DataModel"}
        mocked_spec.protobuf_snippets = invalid_protobuf_snippet_1

        (
            invalid_result_1,
            invalid_msg_1,
        ) = _validate_protocol_buffer_schema_code_snippets(
            mocked_spec, invalid_all_content_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Extra protobuf code snippet provided. Type 'ct:Query' is not used anywhere in your protocol definition."
        )

        invalid_protobuf_snippet_2 = {
            "ct:DataModel": "bytes bytes_field = 1;\nint32 int_field = 2;\nfloat float_field = 3;",
        }
        invalid_all_content_2 = {"ct:DataModel", "ct:Frame"}
        mocked_spec.protobuf_snippets = invalid_protobuf_snippet_2

        (
            invalid_result_2,
            invalid_msg_2,
        ) = _validate_protocol_buffer_schema_code_snippets(
            mocked_spec, invalid_all_content_2
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "No protobuf code snippet is provided for the following custom types: {}".format(
                {"ct:Frame"},
            )
        )

    def test_validate_initiation(self):
        """Test for the '_validate_initiation' method."""
        valid_initiation_1 = ["perm_1", "perm_2"]
        valid_performatives_set_1 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        valid_result_1, valid_msg_1 = _validate_initiation(
            valid_initiation_1, valid_performatives_set_1
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Initial messages are valid."

        ###################################################

        invalid_initiation_1 = []
        invalid_performatives_set_1 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        invalid_result_1, invalid_msg_1 = _validate_initiation(
            invalid_initiation_1, invalid_performatives_set_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "At least one initial performative for this dialogue must be specified."
        )

        invalid_initiation_2 = ["perm_5"]
        invalid_performatives_set_2 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        invalid_result_2, invalid_msg_2 = _validate_initiation(
            invalid_initiation_2, invalid_performatives_set_2
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Performative 'perm_5' specified in \"initiation\" is not defined in the protocol's speech-acts."
        )

    def test_validate_reply(self):
        """Test for the '_validate_reply' method."""
        valid_reply_1 = {
            "performative_ct": ["performative_pct"],
            "performative_pt": ["performative_pmt"],
            "performative_pct": ["performative_mt", "performative_o"],
            "performative_pmt": ["performative_mt", "performative_o"],
            "performative_mt": [],
            "performative_o": [],
            "performative_empty_contents": ["performative_empty_contents"],
        }
        valid_performatives_set_1 = {
            "performative_ct",
            "performative_pt",
            "performative_pct",
            "performative_pmt",
            "performative_mt",
            "performative_o",
            "performative_empty_contents",
        }

        valid_result_1, valid_msg_1, = _validate_reply(
            valid_reply_1, valid_performatives_set_1
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Reply structure is valid."

        ###################################################

        invalid_reply_1 = {
            "perm_1": ["perm_2"],
            "perm_2": ["perm_3"],
            "perm_3": ["perm_4"],
            "perm_4": [],
        }
        invalid_performatives_set_1 = {"perm_1", "perm_2", "perm_3", "perm_4", "perm_5"}

        invalid_result_1, invalid_msg_1, = _validate_reply(
            invalid_reply_1, invalid_performatives_set_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "No reply is provided for the following performatives: {}".format(
                {"perm_5"},
            )
        )

        invalid_reply_2 = {
            "perm_1": ["perm_2"],
            "perm_2": ["perm_3"],
            "perm_3": ["perm_4"],
            "perm_4": ["perm_5"],
            "perm_5": [],
        }
        invalid_performatives_set_2 = {"perm_1", "perm_2", "perm_3", "perm_4"}

        invalid_result_2, invalid_msg_2, = _validate_reply(
            invalid_reply_2, invalid_performatives_set_2
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Performative 'perm_5' specified in \"reply\" is not defined in the protocol's speech-acts."
        )

    def test_validate_termination(self):
        """Test for the '_validate_termination' method."""
        valid_termination_1 = ["perm_4", "perm_3"]
        valid_performatives_set_1 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        valid_result_1, valid_msg_1 = _validate_termination(
            valid_termination_1, valid_performatives_set_1
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Terminal messages are valid."

        ###################################################

        invalid_termination_1 = []
        invalid_performatives_set_1 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        invalid_result_1, invalid_msg_1 = _validate_termination(
            invalid_termination_1, invalid_performatives_set_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "At least one terminal performative for this dialogue must be specified."
        )

        invalid_termination_2 = ["perm_5"]
        invalid_performatives_set_2 = {"perm_1", "perm_2", "perm_3", "perm_4"}
        invalid_result_2, invalid_msg_2 = _validate_termination(
            invalid_termination_2, invalid_performatives_set_2
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Performative 'perm_5' specified in \"termination\" is not defined in the protocol's speech-acts."
        )

    def test_validate_roles(self):
        """Test for the '_validate_roles' method."""
        valid_roles_1 = {"role_1", "role_2"}
        valid_result_1, valid_msg_1 = _validate_roles(valid_roles_1)
        assert valid_result_1 is True
        assert valid_msg_1 == "Dialogue roles are valid."

        valid_roles_2 = {"role_1"}
        valid_result_2, valid_msg_2 = _validate_roles(valid_roles_2)
        assert valid_result_2 is True
        assert valid_msg_2 == "Dialogue roles are valid."

        ###################################################

        invalid_roles_1 = set()
        invalid_result_1, invalid_msg_1 = _validate_roles(invalid_roles_1)
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "There must be either 1 or 2 roles defined in this dialogue. Found 0"
        )

        invalid_roles_2 = {"role_1", "role_2", "role_3"}
        invalid_result_2, invalid_msg_2 = _validate_roles(invalid_roles_2)
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "There must be either 1 or 2 roles defined in this dialogue. Found 3"
        )

        invalid_roles_3 = {"_agent_"}
        invalid_result_3, invalid_msg_3 = _validate_roles(invalid_roles_3)
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Invalid name for role '_agent_'. Role names must match the following regular expression: {} ".format(
                ROLE_REGEX_PATTERN
            )
        )

    def test_validate_end_states(self):
        """Test for the '_validate_end_states' method."""
        valid_end_states_1 = ["end_state_1", "end_state_2"]
        valid_result_1, valid_msg_1 = _validate_end_states(valid_end_states_1)
        assert valid_result_1 is True
        assert valid_msg_1 == "Dialogue end_states are valid."

        valid_end_states_2 = []
        valid_result_2, valid_msg_2 = _validate_end_states(valid_end_states_2)
        assert valid_result_2 is True
        assert valid_msg_2 == "Dialogue end_states are valid."

        ###################################################

        invalid_end_states_1 = ["_end_state_1"]
        invalid_result_1, invalid_msg_1 = _validate_end_states(invalid_end_states_1)
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Invalid name for end_state '_end_state_1'. End_state names must match the following regular expression: {} ".format(
                END_STATE_REGEX_PATTERN
            )
        )

        invalid_end_states_2 = ["end_$tate_1"]
        invalid_result_2, invalid_msg_2 = _validate_end_states(invalid_end_states_2)
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "Invalid name for end_state 'end_$tate_1'. End_state names must match the following regular expression: {} ".format(
                END_STATE_REGEX_PATTERN
            )
        )

    @mock.patch("aea.configurations.base.ProtocolSpecification",)
    def test_validate_dialogue_section(self, mocked_spec):
        """Test for the '_validate_dialogue_section' method."""
        valid_dialogue_config_1 = {
            "initiation": ["performative_ct", "performative_pt"],
            "reply": {
                "performative_ct": ["performative_pct"],
                "performative_pt": ["performative_pmt"],
                "performative_pct": ["performative_mt", "performative_o"],
                "performative_pmt": ["performative_mt", "performative_o"],
                "performative_mt": [],
                "performative_o": [],
                "performative_empty_contents": ["performative_empty_contents"],
            },
            "termination": [
                "performative_mt",
                "performative_o",
                "performative_empty_contents",
            ],
            "roles": {"role_1": None, "role_2": None},
            "end_states": ["end_state_1", "end_state_2", "end_state_3"],
        }
        valid_performatives_set_1 = {
            "performative_ct",
            "performative_pt",
            "performative_pct",
            "performative_pmt",
            "performative_mt",
            "performative_o",
            "performative_empty_contents",
        }
        mocked_spec.dialogue_config = valid_dialogue_config_1

        valid_result_1, valid_msg_1, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert valid_result_1 is True
        assert valid_msg_1 == "Dialogue section of the protocol specification is valid."

        ###################################################

        invalid_dialogue_config_1 = valid_dialogue_config_1.copy()
        invalid_dialogue_config_1["initiation"] = ["new_performative"]

        mocked_spec.dialogue_config = invalid_dialogue_config_1

        invalid_result_1, invalid_msg_1, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert invalid_result_1 is False
        assert (
            invalid_msg_1
            == "Performative 'new_performative' specified in \"initiation\" is not defined in the protocol's speech-acts."
        )

        invalid_dialogue_config_2 = valid_dialogue_config_1.copy()
        invalid_dialogue_config_2["reply"] = {
            "performative_ct": ["performative_pct"],
            "performative_pt": ["performative_pmt"],
            "performative_pct": ["performative_mt", "performative_o"],
            "performative_pmt": ["performative_mt", "performative_o"],
            "performative_mt": [],
            "performative_o": [],
        }

        mocked_spec.dialogue_config = invalid_dialogue_config_2

        invalid_result_2, invalid_msg_2, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert invalid_result_2 is False
        assert (
            invalid_msg_2
            == "No reply is provided for the following performatives: {}".format(
                {"performative_empty_contents"},
            )
        )

        invalid_dialogue_config_3 = valid_dialogue_config_1.copy()
        invalid_dialogue_config_3["termination"] = ["new_performative"]

        mocked_spec.dialogue_config = invalid_dialogue_config_3

        invalid_result_3, invalid_msg_3, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert invalid_result_3 is False
        assert (
            invalid_msg_3
            == "Performative 'new_performative' specified in \"termination\" is not defined in the protocol's speech-acts."
        )

        invalid_dialogue_config_4 = valid_dialogue_config_1.copy()
        invalid_dialogue_config_4["roles"] = {
            "role_1": None,
            "role_2": None,
            "role_3": None,
        }

        mocked_spec.dialogue_config = invalid_dialogue_config_4

        invalid_result_4, invalid_msg_4, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert invalid_result_4 is False
        assert (
            invalid_msg_4
            == "There must be either 1 or 2 roles defined in this dialogue. Found 3"
        )

        invalid_dialogue_config_5 = valid_dialogue_config_1.copy()
        invalid_dialogue_config_5["end_states"] = ["end_$tate_1"]

        mocked_spec.dialogue_config = invalid_dialogue_config_5

        invalid_result_5, invalid_msg_5, = _validate_dialogue_section(
            mocked_spec, valid_performatives_set_1
        )
        assert invalid_result_5 is False
        assert (
            invalid_msg_5
            == "Invalid name for end_state 'end_$tate_1'. End_state names must match the following regular expression: {} ".format(
                END_STATE_REGEX_PATTERN
            )
        )

    @mock.patch("aea.configurations.base.ProtocolSpecification")
    @mock.patch(
        "aea.protocols.generator.validate._validate_speech_acts_section",
        return_value=tuple([True, "Speech_acts are correct!", set(), set()]),
    )
    @mock.patch(
        "aea.protocols.generator.validate._validate_protocol_buffer_schema_code_snippets",
        return_value=tuple([True, "Protobuf snippets are correct!"]),
    )
    @mock.patch(
        "aea.protocols.generator.validate._validate_dialogue_section",
        return_value=tuple([True, "Dialogue section is correct!"]),
    )
    def test_validate_positive(
        self,
        mocked_spec,
        macked_validate_speech_acts,
        macked_validate_protobuf,
        macked_validate_dialogue,
    ):
        """Positive test for the 'validate' method: invalid dialogue section."""
        valid_result_1, valid_msg_1, = validate(mocked_spec)
        assert valid_result_1 is True
        assert valid_msg_1 == "Protocol specification is valid."

    @mock.patch("aea.configurations.base.ProtocolSpecification")
    @mock.patch(
        "aea.protocols.generator.validate._validate_speech_acts_section",
        return_value=tuple([False, "Some error on speech_acts.", None, None]),
    )
    def test_validate_negative_invalid_speech_acts(
        self, mocked_spec, macked_validate_speech_acts
    ):
        """Negative test for the 'validate' method: invalid speech_acts."""
        invalid_result_1, invalid_msg_1, = validate(mocked_spec)
        assert invalid_result_1 is False
        assert invalid_msg_1 == "Some error on speech_acts."

    @mock.patch("aea.configurations.base.ProtocolSpecification")
    @mock.patch(
        "aea.protocols.generator.validate._validate_speech_acts_section",
        return_value=tuple([True, "Speech_acts are correct!", set(), set()]),
    )
    @mock.patch(
        "aea.protocols.generator.validate._validate_protocol_buffer_schema_code_snippets",
        return_value=tuple([False, "Some error on protobuf snippets."]),
    )
    def test_validate_negative_invalid_protobuf_snippets(
        self, mocked_spec, macked_validate_speech_acts, macked_validate_protobuf
    ):
        """Negative test for the 'validate' method: invalid protobuf snippets."""
        invalid_result_1, invalid_msg_1, = validate(mocked_spec)
        assert invalid_result_1 is False
        assert invalid_msg_1 == "Some error on protobuf snippets."

    @mock.patch("aea.configurations.base.ProtocolSpecification")
    @mock.patch(
        "aea.protocols.generator.validate._validate_speech_acts_section",
        return_value=tuple([True, "Speech_acts are correct!", set(), set()]),
    )
    @mock.patch(
        "aea.protocols.generator.validate._validate_protocol_buffer_schema_code_snippets",
        return_value=tuple([True, "Protobuf snippets are correct!"]),
    )
    @mock.patch(
        "aea.protocols.generator.validate._validate_dialogue_section",
        return_value=tuple([False, "Some error on dialogue section."]),
    )
    def test_validate_negative_invalid_dialogue_section(
        self,
        mocked_spec,
        macked_validate_speech_acts,
        macked_validate_protobuf,
        macked_validate_dialogue,
    ):
        """Negative test for the 'validate' method: invalid dialogue section."""
        invalid_result_1, invalid_msg_1, = validate(mocked_spec)
        assert invalid_result_1 is False
        assert invalid_msg_1 == "Some error on dialogue section."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2020 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for signing protocol."""

from typing import Any, Dict, cast

from aea.protocols.base import Message, Serializer
from aea.protocols.signing import signing_pb2
from aea.protocols.signing.custom_types import (
    ErrorCode,
    RawMessage,
    RawTransaction,
    SignedMessage,
    SignedTransaction,
    Terms,
)
from aea.protocols.signing.message import SigningMessage


class SigningSerializer(Serializer):
    """Serialization for the 'signing' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Signing' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(SigningMessage, msg)
        signing_msg = signing_pb2.SigningMessage()
        signing_msg.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        signing_msg.dialogue_starter_reference = dialogue_reference[0]
        signing_msg.dialogue_responder_reference = dialogue_reference[1]
        signing_msg.target = msg.target

        performative_id = msg.performative
        if performative_id == SigningMessage.Performative.SIGN_TRANSACTION:
            performative = signing_pb2.SigningMessage.Sign_Transaction_Performative()  # type: ignore
            terms = msg.terms
            Terms.encode(performative.terms, terms)
            raw_transaction = msg.raw_transaction
            RawTransaction.encode(performative.raw_transaction, raw_transaction)
            signing_msg.sign_transaction.CopyFrom(performative)
        elif performative_id == SigningMessage.Performative.SIGN_MESSAGE:
            performative = signing_pb2.SigningMessage.Sign_Message_Performative()  # type: ignore
            terms = msg.terms
            Terms.encode(performative.terms, terms)
            raw_message = msg.raw_message
            RawMessage.encode(performative.raw_message, raw_message)
            signing_msg.sign_message.CopyFrom(performative)
        elif performative_id == SigningMessage.Performative.SIGNED_TRANSACTION:
            performative = signing_pb2.SigningMessage.Signed_Transaction_Performative()  # type: ignore
            signed_transaction = msg.signed_transaction
            SignedTransaction.encode(
                performative.signed_transaction, signed_transaction
            )
            signing_msg.signed_transaction.CopyFrom(performative)
        elif performative_id == SigningMessage.Performative.SIGNED_MESSAGE:
            performative = signing_pb2.SigningMessage.Signed_Message_Performative()  # type: ignore
            signed_message = msg.signed_message
            SignedMessage.encode(performative.signed_message, signed_message)
            signing_msg.signed_message.CopyFrom(performative)
        elif performative_id == SigningMessage.Performative.ERROR:
            performative = signing_pb2.SigningMessage.Error_Performative()  # type: ignore
            error_code = msg.error_code
            ErrorCode.encode(performative.error_code, error_code)
            signing_msg.error.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        signing_bytes = signing_msg.SerializeToString()
        return signing_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Signing' message.

        :param obj: the bytes object.
        :return: the 'Signing' message.
        """
        signing_pb = signing_pb2.SigningMessage()
        signing_pb.ParseFromString(obj)
        message_id = signing_pb.message_id
        dialogue_reference = (
            signing_pb.dialogue_starter_reference,
            signing_pb.dialogue_responder_reference,
        )
        target = signing_pb.target

        performative = signing_pb.WhichOneof("performative")
        performative_id = SigningMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == SigningMessage.Performative.SIGN_TRANSACTION:
            pb2_terms = signing_pb.sign_transaction.terms
            terms = Terms.decode(pb2_terms)
            performative_content["terms"] = terms
            pb2_raw_transaction = signing_pb.sign_transaction.raw_transaction
            raw_transaction = RawTransaction.decode(pb2_raw_transaction)
            performative_content["raw_transaction"] = raw_transaction
        elif performative_id == SigningMessage.Performative.SIGN_MESSAGE:
            pb2_terms = signing_pb.sign_message.terms
            terms = Terms.decode(pb2_terms)
            performative_content["terms"] = terms
            pb2_raw_message = signing_pb.sign_message.raw_message
            raw_message = RawMessage.decode(pb2_raw_message)
            performative_content["raw_message"] = raw_message
        elif performative_id == SigningMessage.Performative.SIGNED_TRANSACTION:
            pb2_signed_transaction = signing_pb.signed_transaction.signed_transaction
            signed_transaction = SignedTransaction.decode(pb2_signed_transaction)
            performative_content["signed_transaction"] = signed_transaction
        elif performative_id == SigningMessage.Performative.SIGNED_MESSAGE:
            pb2_signed_message = signing_pb.signed_message.signed_message
            signed_message = SignedMessage.decode(pb2_signed_message)
            performative_content["signed_message"] = signed_message
        elif performative_id == SigningMessage.Performative.ERROR:
            pb2_error_code = signing_pb.error.error_code
            error_code = ErrorCode.decode(pb2_error_code)
            performative_content["error_code"] = error_code
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return SigningMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the http protocol package."""

import sys
from typing import Type
from unittest import mock

import pytest

from aea.common import Address
from aea.exceptions import AEAEnforceError
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

import packages
from packages.fetchai.protocols.tac.dialogues import TacDialogue, TacDialogues
from packages.fetchai.protocols.tac.message import TacMessage
from packages.fetchai.protocols.tac.message import _default_logger as tac_message_logger

from tests.conftest import ROOT_DIR


sys.path.append(ROOT_DIR)


def test_tac_message_instantiation():
    """Test instantiation of the tac message."""
    assert TacMessage(
        performative=TacMessage.Performative.REGISTER, agent_name="some_name"
    )
    assert TacMessage(performative=TacMessage.Performative.UNREGISTER)
    assert TacMessage(
        performative=TacMessage.Performative.TRANSACTION,
        transaction_id="some_id",
        ledger_id="some_ledger",
        sender_address="some_address",
        counterparty_address="some_other_address",
        amount_by_currency_id={"FET": 10},
        fee_by_currency_id={"FET": 1},
        quantities_by_good_id={"123": 0, "1234": 10},
        nonce=1,
        sender_signature="some_signature",
        counterparty_signature="some_other_signature",
    )
    assert TacMessage(performative=TacMessage.Performative.CANCELLED)
    assert TacMessage(
        performative=TacMessage.Performative.GAME_DATA,
        amount_by_currency_id={"FET": 10},
        exchange_params_by_currency_id={"FET": 10.0},
        quantities_by_good_id={"123": 20, "1234": 15},
        utility_params_by_good_id={"123": 30.0, "1234": 50.0},
        fee_by_currency_id={"FET": 1},
        agent_addr_to_name={"agent_1": "Agent one", "agent_2": "Agent two"},
        currency_id_to_name={"FET": "currency_name"},
        good_id_to_name={"123": "First good", "1234": "Second good"},
        version_id="game_version_1",
    )
    assert TacMessage(
        performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
        transaction_id="some_id",
        amount_by_currency_id={"FET": 10},
        quantities_by_good_id={"123": 20, "1234": 15},
    )
    assert TacMessage(
        performative=TacMessage.Performative.TAC_ERROR,
        error_code=TacMessage.ErrorCode.GENERIC_ERROR,
        info={"msg": "This is info msg."},
    )
    assert str(TacMessage.Performative.REGISTER) == "register"


def test_register_serialization():
    """Test the serialization for 'register' speech-act works."""
    msg = TacMessage(
        performative=TacMessage.Performative.REGISTER, agent_name="some_agent_name",
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_unregister_serialization():
    """Test the serialization for 'unregister' speech-act works."""
    msg = TacMessage(
        message_id=2, target=1, performative=TacMessage.Performative.UNREGISTER,
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_transaction_serialization():
    """Test the serialization for 'transaction' speech-act works."""
    msg = TacMessage(
        performative=TacMessage.Performative.TRANSACTION,
        transaction_id="some_transaction_id",
        ledger_id="some_ledger_id",
        sender_address="some_sender_address",
        counterparty_address="some_counterparty_address",
        amount_by_currency_id={"key_1": 1, "key_2": 2},
        fee_by_currency_id={"key_1": 1, "key_2": 2},
        quantities_by_good_id={"key_1": 1, "key_2": 2},
        nonce="some_nonce",
        sender_signature="some_sender_signature",
        counterparty_signature="some_counterparty_signature",
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_cancelled_serialization():
    """Test the serialization for 'cancelled' speech-act works."""
    msg = TacMessage(performative=TacMessage.Performative.CANCELLED,)
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_game_data_serialization():
    """Test the serialization for 'game_data' speech-act works."""
    msg = TacMessage(
        performative=TacMessage.Performative.GAME_DATA,
        amount_by_currency_id={"key_1": 1, "key_2": 2},
        exchange_params_by_currency_id={"key_1": 1.0, "key_2": 2.0},
        quantities_by_good_id={"key_1": 1, "key_2": 2},
        utility_params_by_good_id={"key_1": 1.0, "key_2": 2.0},
        fee_by_currency_id={"key_1": 1, "key_2": 2},
        agent_addr_to_name={"key_1": "value_1", "key_2": "value_2"},
        currency_id_to_name={"key_1": "value_1", "key_2": "value_2"},
        good_id_to_name={"key_1": "value_1", "key_2": "value_2"},
        version_id="some_version_id",
        info={"key_1": "value_1", "key_2": "value_2"},
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_transaction_confirmation_serialization():
    """Test the serialization for 'transaction_confirmation' speech-act works."""
    msg = TacMessage(
        performative=TacMessage.Performative.TRANSACTION_CONFIRMATION,
        transaction_id="some_transaction_id",
        amount_by_currency_id={"key_1": 1, "key_2": 2},
        quantities_by_good_id={"key_1": 1, "key_2": 2},
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_tac_error_serialization():
    """Test the serialization for 'tac_error' speech-act works."""
    msg = TacMessage(
        performative=TacMessage.Performative.TAC_ERROR,
        error_code=TacMessage.ErrorCode.GENERIC_ERROR,
        info={"key_1": "value_1", "key_2": "value_2"},
    )
    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to, sender="sender", protocol_id=TacMessage.protocol_id, message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert expected_envelope.protocol_id == actual_envelope.protocol_id
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TacMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_oef_type_string_value():
    """Test the string value of the type."""
    assert (
        str(TacMessage.Performative.REGISTER) == "register"
    ), "The str value must be register"
    assert (
        str(TacMessage.Performative.UNREGISTER) == "unregister"
    ), "The str value must be unregister"
    assert (
        str(TacMessage.Performative.TRANSACTION) == "transaction"
    ), "The str value must be transaction"
    assert (
        str(TacMessage.Performative.CANCELLED) == "cancelled"
    ), "The str value must be cancelled"
    assert (
        str(TacMessage.Performative.GAME_DATA) == "game_data"
    ), "The str value must be game_data"
    assert (
        str(TacMessage.Performative.TRANSACTION_CONFIRMATION)
        == "transaction_confirmation"
    ), "The str value must be transaction_confirmation"
    assert (
        str(TacMessage.Performative.TAC_ERROR) == "tac_error"
    ), "The str value must be tac_error"


def test_error_code_to_msg():
    """Test the serialization for 'tac_error' speech-act works."""

    assert (
        str(TacMessage.ErrorCode.to_msg(0)) == "Unexpected error."
    ), 'The str value must be "Unexpected error."'
    assert (
        str(TacMessage.ErrorCode.to_msg(1)) == "Request not recognized"
    ), 'The str value must be "Request not recognized"'
    assert (
        str(TacMessage.ErrorCode.to_msg(2)) == "Agent addr already registered."
    ), 'The str value must be "Agent addr already registered."'
    assert (
        str(TacMessage.ErrorCode.to_msg(3)) == "Agent name already registered."
    ), 'The str value must be "Agent name already registered."'
    assert (
        str(TacMessage.ErrorCode.to_msg(4)) == "Agent not registered."
    ), 'The str value must be "Agent not registered."'
    assert (
        str(TacMessage.ErrorCode.to_msg(5)) == "Error in checking transaction"
    ), 'The str value must be "Error in checking transaction"'
    assert (
        str(TacMessage.ErrorCode.to_msg(6))
        == "The transaction request does not match with a previous transaction request with the same id."
    ), 'The str value must be "The transaction request does not match with a previous transaction request with the same id."'
    assert (
        str(TacMessage.ErrorCode.to_msg(7)) == "Agent name not in whitelist."
    ), 'The str value must be "Agent name not in whitelist."'
    assert (
        str(TacMessage.ErrorCode.to_msg(8)) == "The competition is not running yet."
    ), 'The str value must be "The competition is not running yet."'
    assert (
        str(TacMessage.ErrorCode.to_msg(9))
        == "The message is inconsistent with the dialogue."
    ), 'The str value must be "The message is inconsistent with the dialogue."'


def test_encoding_unknown_performative():
    """Test that we raise an exception when the performative is unknown during encoding."""
    msg = TacMessage(performative=TacMessage.Performative.CANCELLED,)

    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(TacMessage.Performative, "__eq__", return_value=False):
            TacMessage.serializer.encode(msg)


def test_decoding_unknown_performative():
    """Test that we raise an exception when the performative is unknown during decoding."""
    msg = TacMessage(performative=TacMessage.Performative.CANCELLED,)

    encoded_msg = TacMessage.serializer.encode(msg)
    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(TacMessage.Performative, "__eq__", return_value=False):
            TacMessage.serializer.decode(encoded_msg)


@mock.patch.object(
    packages.fetchai.protocols.tac.message,
    "enforce",
    side_effect=AEAEnforceError("some error"),
)
def test_incorrect_message(mocked_enforce):
    """Test that we raise an exception when the message is incorrect."""
    with mock.patch.object(tac_message_logger, "error") as mock_logger:
        TacMessage(performative=TacMessage.Performative.CANCELLED,)

        mock_logger.assert_any_call("some error")


class TestDialogues:
    """Tests tac dialogues."""

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        cls.agent_addr = "agent address"
        cls.controller_addr = "controller address"
        cls.agent_dialogues = AgentDialogues(cls.agent_addr)
        cls.controller_dialogues = ControllerDialogues(cls.controller_addr)

    def test_create_self_initiated(self):
        """Test the self initialisation of a dialogue."""
        result = self.agent_dialogues._create_self_initiated(
            dialogue_opponent_addr=self.controller_addr,
            dialogue_reference=(str(0), ""),
            role=TacDialogue.Role.PARTICIPANT,
        )
        assert isinstance(result, TacDialogue)
        assert (
            result.role == TacDialogue.Role.PARTICIPANT
        ), "The role must be participant."

    def test_create_opponent_initiated(self):
        """Test the opponent initialisation of a dialogue."""
        result = self.agent_dialogues._create_opponent_initiated(
            dialogue_opponent_addr=self.controller_addr,
            dialogue_reference=(str(0), ""),
            role=TacDialogue.Role.PARTICIPANT,
        )
        assert isinstance(result, TacDialogue)
        assert (
            result.role == TacDialogue.Role.PARTICIPANT
        ), "The role must be participant."


class AgentDialogue(TacDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[TacMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for

        :return: None
        """
        TacDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class AgentDialogues(TacDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return TacDialogue.Role.PARTICIPANT

        TacDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=AgentDialogue,
        )


class ControllerDialogue(TacDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[TacMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for

        :return: None
        """
        TacDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class ControllerDialogues(TacDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return TacDialogue.Role.CONTROLLER

        TacDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=ControllerDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Tests for the pipe module."""

import asyncio
from threading import Thread

import pytest

from aea.helpers.pipe import (
    IPCChannelClient,
    PosixNamedPipeChannel,
    PosixNamedPipeChannelClient,
    TCPSocketChannel,
    TCPSocketChannelClient,
    make_ipc_channel,
    make_ipc_channel_client,
)

from tests.conftest import skip_test_windows


def _run_echo_service(client: IPCChannelClient):
    async def echo_service(client: IPCChannelClient):
        try:
            await client.connect()
            while True:
                data = await client.read()
                if not data:
                    break
                await client.write(data)
        except (asyncio.IncompleteReadError, asyncio.CancelledError, OSError):
            pass
        finally:
            await client.close()

    loop = asyncio.new_event_loop()
    loop.run_until_complete(echo_service(client))


@pytest.mark.asyncio
class TestAEAHelperMakePipe:
    """Test that make_ipc_channel utility and abstract class IPCChannel work properly"""

    @pytest.mark.asyncio
    async def test_connection_communication(self):
        """Test connection communication."""
        pipe = make_ipc_channel()
        assert (
            pipe.in_path is not None and pipe.out_path is not None
        ), "Pipe not properly setup"

        connected = asyncio.ensure_future(pipe.connect())

        client_pipe = make_ipc_channel_client(pipe.out_path, pipe.in_path)

        client = Thread(target=_run_echo_service, args=[client_pipe])
        client.start()

        try:
            assert await connected, "Failed to connect pipe"

            message = b"hello"
            await pipe.write(message)
            received = await pipe.read()

            assert received == message, "Echoed message differs"

        except Exception:
            raise
        finally:
            await pipe.close()
            client.join()


@pytest.mark.asyncio
class TestAEAHelperTCPSocketChannel:
    """Test that TCPSocketChannel work properly"""

    @pytest.mark.asyncio
    async def test_connection_communication(self):
        """Test connection communication."""
        pipe = TCPSocketChannel()
        assert (
            pipe.in_path is not None and pipe.out_path is not None
        ), "TCPSocketChannel not properly setup"

        connected = asyncio.ensure_future(pipe.connect())

        client_pipe = TCPSocketChannelClient(pipe.out_path, pipe.in_path)

        client = Thread(target=_run_echo_service, args=[client_pipe])
        client.start()

        try:
            assert await connected, "Failed to connect pipe"

            message = b"hello"
            await pipe.write(message)
            received = await pipe.read()

            assert received == message, "Echoed message differs"

        except Exception:
            raise
        finally:
            await pipe.close()
            client.join()

    @pytest.mark.asyncio
    async def test_connection_refused(self):
        """Test connection refused."""
        pipe = TCPSocketChannel()
        assert (
            pipe.in_path is not None and pipe.out_path is not None
        ), "TCPSocketChannel not properly setup"

        client_pipe = TCPSocketChannelClient(pipe.out_path, pipe.in_path)

        connected = await client_pipe.connect()
        assert connected is False


@skip_test_windows
@pytest.mark.asyncio
class TestAEAHelperPosixNamedPipeChannel:
    """Test that TCPSocketChannel work properly"""

    @pytest.mark.asyncio
    async def test_connection_communication(self):
        """Test connection communication."""
        pipe = PosixNamedPipeChannel()
        assert (
            pipe.in_path is not None and pipe.out_path is not None
        ), "PosixNamedPipeChannel not properly setup"

        connected = asyncio.ensure_future(pipe.connect())

        client_pipe = PosixNamedPipeChannelClient(pipe.out_path, pipe.in_path)

        client = Thread(target=_run_echo_service, args=[client_pipe])
        client.start()

        try:
            assert await connected, "Failed to connect pipe"

            message = b"hello"
            await pipe.write(message)
            received = await pipe.read()

            assert received == message, "Echoed message differs"

        except Exception:
            raise
        finally:
            await pipe.close()
            client.join()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the stub connection."""
import asyncio
import base64
import os
import shutil
import tempfile
import time
from pathlib import Path
from unittest import mock

import pytest

import aea
from aea.configurations.base import PublicId
from aea.connections.stub.connection import (
    StubConnection,
    _process_line,
    lock_file,
    write_envelope,
    write_with_lock,
)
from aea.crypto.wallet import CryptoStore
from aea.identity.base import Identity
from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer
from aea.protocols.default.message import DefaultMessage

from tests.conftest import ROOT_DIR, _make_stub_connection


SEPARATOR = ","


def make_test_envelope() -> Envelope:
    """Create a test envelope."""
    msg = DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"hello",
    )
    msg.to = "any"
    envelope = Envelope(
        to="any", sender="any", protocol_id=DefaultMessage.protocol_id, message=msg,
    )
    return envelope


class TestStubConnectionReception:
    """Test that the stub connection is implemented correctly."""

    def setup(self):
        """Set the test up."""
        self.cwd = os.getcwd()
        self.tmpdir = Path(tempfile.mkdtemp())
        d = self.tmpdir / "test_stub"
        d.mkdir(parents=True)
        self.input_file_path = d / "input_file.csv"
        self.output_file_path = d / "output_file.csv"
        self.connection = _make_stub_connection(
            self.input_file_path, self.output_file_path
        )

        self.multiplexer = Multiplexer([self.connection])
        self.multiplexer.connect()
        os.chdir(self.tmpdir)

    def test_reception_a(self):
        """Test that the connection receives what has been enqueued in the input file."""
        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        expected_envelope = Envelope(
            to="any",
            sender="anys",
            protocol_id=DefaultMessage.protocol_id,
            message=msg,
        )

        with open(self.input_file_path, "ab+") as f:
            write_envelope(expected_envelope, f)

        actual_envelope = self.multiplexer.get(block=True, timeout=3.0)
        assert expected_envelope.to == actual_envelope.to
        assert expected_envelope.sender == actual_envelope.sender
        assert expected_envelope.protocol_id == actual_envelope.protocol_id
        msg = DefaultMessage.serializer.decode(actual_envelope.message)
        msg.to = actual_envelope.to
        msg.sender = actual_envelope.sender
        assert expected_envelope.message == msg

    def test_reception_b(self):
        """Test that the connection receives what has been enqueued in the input file."""
        # a message containing delimiters and newline characters
        msg = b"\x08\x02\x12\x011\x1a\x011 \x01:,\n*0x32468d\n,\nB8Ab795\n\n49B49C88DC991990E7910891,,dbd\n"
        protocol_id = PublicId.from_str("some_author/some_name:0.1.0")
        encoded_envelope = "{}{}{}{}{}{}{}{}".format(
            "any",
            SEPARATOR,
            "any",
            SEPARATOR,
            protocol_id,
            SEPARATOR,
            msg.decode("utf-8"),
            SEPARATOR,
        )
        encoded_envelope = encoded_envelope.encode("utf-8")

        with open(self.input_file_path, "ab+") as f:
            write_with_lock(f, encoded_envelope)

        actual_envelope = self.multiplexer.get(block=True, timeout=3.0)
        assert "any" == actual_envelope.to
        assert "any" == actual_envelope.sender
        assert protocol_id == actual_envelope.protocol_id
        assert msg == actual_envelope.message

    def test_reception_c(self):
        """Test that the connection receives what has been enqueued in the input file."""
        encoded_envelope = b"0x5E22777dD831A459535AA4306AceC9cb22eC4cB5,default_oef,fetchai/oef_search:0.8.0,\x08\x02\x12\x011\x1a\x011 \x01:,\n*0x32468dB8Ab79549B49C88DC991990E7910891dbd,"
        expected_envelope = Envelope(
            to="0x5E22777dD831A459535AA4306AceC9cb22eC4cB5",
            sender="default_oef",
            protocol_id=PublicId.from_str("fetchai/oef_search:0.8.0"),
            message=b"\x08\x02\x12\x011\x1a\x011 \x01:,\n*0x32468dB8Ab79549B49C88DC991990E7910891dbd",
        )
        with open(self.input_file_path, "ab+") as f:
            write_with_lock(f, encoded_envelope)

        actual_envelope = self.multiplexer.get(block=True, timeout=3.0)
        assert expected_envelope == actual_envelope

    def test_reception_fails(self):
        """Test the case when an error occurs during the processing of a line."""
        patch = mock.patch.object(
            aea.connections.stub.connection._default_logger, "error"
        )
        mocked_logger_error = patch.start()
        with mock.patch(
            "aea.connections.stub.connection._decode",
            side_effect=Exception("an error."),
        ):
            _process_line(b"")
            mocked_logger_error.assert_called_with(
                "Error when processing a line. Message: an error.", exc_info=True
            )

        patch.stop()

    def teardown(self):
        """Tear down the test."""
        os.chdir(self.cwd)
        try:
            shutil.rmtree(self.tmpdir)
        except (OSError, IOError):
            pass
        self.multiplexer.disconnect()


class TestStubConnectionSending:
    """Test that the stub connection is implemented correctly."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.cwd = os.getcwd()
        cls.tmpdir = Path(tempfile.mkdtemp())
        d = cls.tmpdir / "test_stub"
        d.mkdir(parents=True)
        cls.input_file_path = d / "input_file.csv"
        cls.output_file_path = d / "output_file.csv"
        cls.connection = _make_stub_connection(
            cls.input_file_path, cls.output_file_path
        )

        cls.multiplexer = Multiplexer([cls.connection])
        cls.multiplexer.connect()
        os.chdir(cls.tmpdir)

    def test_connection_is_established(self):
        """Test the stub connection is established and then bad formatted messages."""
        assert self.connection.is_connected
        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        encoded_envelope = "{}{}{}{}{}{}{}{}".format(
            "any",
            SEPARATOR,
            "any",
            SEPARATOR,
            DefaultMessage.protocol_id,
            SEPARATOR,
            DefaultMessage.serializer.encode(msg).decode("utf-8"),
            SEPARATOR,
        )
        encoded_envelope = base64.b64encode(encoded_envelope.encode("utf-8"))
        envelope = _process_line(encoded_envelope)
        if envelope is not None:
            self.connection._put_envelopes([envelope])

        assert (
            self.connection.in_queue.empty()
        ), "The inbox must be empty due to bad encoded message"

    def test_send_message(self):
        """Test that the messages in the outbox are posted on the output file."""
        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        expected_envelope = Envelope(
            to="any",
            sender="anys",
            protocol_id=DefaultMessage.protocol_id,
            message=msg,
        )

        self.multiplexer.put(expected_envelope)
        time.sleep(0.1)

        with open(self.output_file_path, "rb+") as f:
            lines = f.readlines()

        assert len(lines) == 2
        line = lines[0] + lines[1]
        to, sender, protocol_id, message, end = line.strip().split(
            "{}".format(SEPARATOR).encode("utf-8"), maxsplit=4
        )
        to = to.decode("utf-8")
        sender = sender.decode("utf-8")
        protocol_id = PublicId.from_str(protocol_id.decode("utf-8"))
        assert end in [b"", b"\n"]

        actual_envelope = Envelope(
            to=to, sender=sender, protocol_id=protocol_id, message=message
        )
        assert expected_envelope.to == actual_envelope.to
        assert expected_envelope.sender == actual_envelope.sender
        assert expected_envelope.protocol_id == actual_envelope.protocol_id
        msg = DefaultMessage.serializer.decode(actual_envelope.message)
        msg.to = actual_envelope.to
        msg.sender = actual_envelope.sender
        assert expected_envelope.message == msg

    @classmethod
    def teardown_class(cls):
        """Tear down the test."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.tmpdir)
        except (OSError, IOError):
            pass
        cls.multiplexer.disconnect()


@pytest.mark.asyncio
async def test_disconnection_when_already_disconnected():
    """Test the case when disconnecting a connection already disconnected."""
    tmpdir = Path(tempfile.mkdtemp())
    d = tmpdir / "test_stub"
    d.mkdir(parents=True)
    input_file_path = d / "input_file.csv"
    output_file_path = d / "output_file.csv"
    connection = _make_stub_connection(input_file_path, output_file_path)

    assert not connection.is_connected
    await connection.disconnect()
    assert not connection.is_connected


@pytest.mark.asyncio
async def test_connection_when_already_connected():
    """Test the case when connecting a connection already connected."""
    tmpdir = Path(tempfile.mkdtemp())
    d = tmpdir / "test_stub"
    d.mkdir(parents=True)
    input_file_path = d / "input_file.csv"
    output_file_path = d / "output_file.csv"
    connection = _make_stub_connection(input_file_path, output_file_path)

    assert not connection.is_connected
    await connection.connect()
    assert connection.is_connected
    await connection.connect()
    assert connection.is_connected

    await connection.disconnect()


@pytest.mark.asyncio
async def test_receiving_returns_none_when_error_occurs():
    """Test that when we try to receive an envelope and an error occurs we return None."""
    tmpdir = Path(tempfile.mkdtemp())
    d = tmpdir / "test_stub"
    d.mkdir(parents=True)
    input_file_path = d / "input_file.csv"
    output_file_path = d / "output_file.csv"
    connection = _make_stub_connection(input_file_path, output_file_path)

    await connection.connect()
    with mock.patch.object(connection.in_queue, "get", side_effect=Exception):
        ret = await connection.receive()
        assert ret is None

    await connection.disconnect()


@pytest.mark.asyncio
async def test_multiple_envelopes():
    """Test many envelopes received."""
    tmpdir = Path(tempfile.mkdtemp())
    d = tmpdir / "test_stub"
    d.mkdir(parents=True)
    input_file_path = d / "input_file.csv"
    output_file_path = d / "output_file.csv"
    connection = _make_stub_connection(input_file_path, output_file_path)

    num_envelopes = 5
    await connection.connect()
    assert connection.is_connected

    async def wait_num(num):
        for _ in range(num):
            assert await connection.receive()

    task = asyncio.get_event_loop().create_task(wait_num(num_envelopes))

    with open(input_file_path, "ab+") as f:
        for _ in range(num_envelopes):
            write_envelope(make_test_envelope(), f)
            await asyncio.sleep(0.01)  # spin asyncio loop

    await asyncio.wait_for(task, timeout=3)
    await connection.disconnect()


@pytest.mark.asyncio
async def test_bad_envelope():
    """Test bad format envelop."""
    tmpdir = Path(tempfile.mkdtemp())
    d = tmpdir / "test_stub"
    d.mkdir(parents=True)
    input_file_path = d / "input_file.csv"
    output_file_path = d / "output_file.csv"
    connection = _make_stub_connection(input_file_path, output_file_path)

    await connection.connect()

    with open(input_file_path, "ab+") as f:
        f.write(b"1,2,3,4,5,")
        f.flush()

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(connection.receive(), timeout=0.1)

    await connection.disconnect()


@pytest.mark.asyncio
async def test_load_from_dir():
    """Test stub connection can be loaded from dir."""
    StubConnection.from_dir(
        ROOT_DIR + "/aea/connections/stub", Identity("name", "address"), CryptoStore(),
    )


class TestFileLock:
    """Test for filelocks."""

    def test_lock_file_ok(self):
        """Work ok ok for random file."""
        with tempfile.TemporaryFile() as fp:
            with lock_file(fp):
                pass

    def test_lock_file_error(self):
        """Fail on closed file."""
        with tempfile.TemporaryFile() as fp:
            fp.close()
            with pytest.raises(ValueError):
                with lock_file(fp):
                    pass



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for Libp2p tcp client connection."""

import os
import shutil
import tempfile

import pytest

from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer
from aea.protocols.default.message import DefaultMessage
from aea.protocols.default.serialization import DefaultSerializer

from packages.fetchai.connections.p2p_libp2p_client.connection import Uri

from tests.conftest import (
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    libp2p_log_on_failure,
    libp2p_log_on_failure_all,
)


DEFAULT_PORT = 10234
DEFAULT_DELEGATE_PORT = 11234
DEFAULT_HOST = "127.0.0.1"
DEFAULT_CLIENTS_PER_NODE = 4


@pytest.mark.asyncio
class TestLibp2pClientConnectionConnectDisconnect:
    """Test that connection is established and torn down correctly"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.connection_node = _make_libp2p_connection(delegate=True)
        cls.connection = _make_libp2p_client_connection()

    @pytest.mark.asyncio
    async def test_libp2pclientconnection_connect_disconnect(self):
        """Test connnect then disconnect."""
        assert self.connection.is_connected is False
        try:
            await self.connection_node.connect()
            await self.connection.connect()
            assert self.connection.is_connected is True

            await self.connection.disconnect()
            assert self.connection.is_connected is False
        except Exception:
            raise
        finally:
            await self.connection_node.disconnect()

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelope:
    """Test that connection will route envelope to destination through the same libp2p node"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []

        cls.connection_node = _make_libp2p_connection(DEFAULT_PORT + 1, delegate=True)
        cls.multiplexer_node = Multiplexer([cls.connection_node])
        cls.log_files.append(cls.connection_node.node.log_file)
        cls.multiplexer_node.connect()

        try:
            cls.connection_client_1 = _make_libp2p_client_connection()
            cls.multiplexer_client_1 = Multiplexer([cls.connection_client_1])
            cls.multiplexer_client_1.connect()

            cls.connection_client_2 = _make_libp2p_client_connection()
            cls.multiplexer_client_2 = Multiplexer([cls.connection_client_2])
            cls.multiplexer_client_2.connect()
        except Exception:
            cls.multiplexer_node.disconnect()
            raise

    def test_connection_is_established(self):
        """Test connection is established."""
        assert self.connection_client_1.is_connected is True
        assert self.connection_client_2.is_connected is True

    def test_envelope_routed(self):
        """Test the envelope is routed."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert delivered_envelope.protocol_id == envelope.protocol_id
        assert delivered_envelope.message == envelope.message

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_2

        self.multiplexer_client_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_n = self.connection_node.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_n,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_node.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_n

        self.multiplexer_node.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        cls.multiplexer_client_1.disconnect()
        cls.multiplexer_client_2.disconnect()
        cls.multiplexer_node.disconnect()

        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionEchoEnvelopeTwoDHTNode:
    """Test that connection will route envelope to destination connected to different node"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []
        cls.mutliplexers = []

        cls.connection_node_1 = _make_libp2p_connection(
            port=DEFAULT_PORT + 1,
            delegate_port=DEFAULT_DELEGATE_PORT + 1,
            delegate=True,
        )
        cls.multiplexer_node_1 = Multiplexer([cls.connection_node_1])
        cls.log_files.append(cls.connection_node_1.node.log_file)
        cls.multiplexer_node_1.connect()
        cls.mutliplexers.append(cls.multiplexer_node_1)

        genesis_peer = cls.connection_node_1.node.multiaddrs[0]

        try:
            cls.connection_node_2 = _make_libp2p_connection(
                port=DEFAULT_PORT + 2,
                delegate_port=DEFAULT_DELEGATE_PORT + 2,
                entry_peers=[genesis_peer],
                delegate=True,
            )
            cls.multiplexer_node_2 = Multiplexer([cls.connection_node_2])
            cls.log_files.append(cls.connection_node_2.node.log_file)
            cls.multiplexer_node_2.connect()
            cls.mutliplexers.append(cls.multiplexer_node_2)

            cls.connection_client_1 = _make_libp2p_client_connection(
                DEFAULT_DELEGATE_PORT + 1
            )
            cls.multiplexer_client_1 = Multiplexer([cls.connection_client_1])
            cls.multiplexer_client_1.connect()
            cls.mutliplexers.append(cls.multiplexer_client_1)

            cls.connection_client_2 = _make_libp2p_client_connection(
                DEFAULT_DELEGATE_PORT + 2
            )
            cls.multiplexer_client_2 = Multiplexer([cls.connection_client_2])
            cls.multiplexer_client_2.connect()
            cls.mutliplexers.append(cls.multiplexer_client_2)
        except Exception:
            cls.teardown_class()
            raise

    def test_connection_is_established(self):
        """Test the connection is established."""
        assert self.connection_node_1.is_connected is True
        assert self.connection_node_2.is_connected is True
        assert self.connection_client_1.is_connected is True
        assert self.connection_client_2.is_connected is True

    def test_envelope_routed(self):
        """Test the envelope is routed."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=20)

        assert delivered_envelope is not None
        assert delivered_envelope.to == envelope.to
        assert delivered_envelope.sender == envelope.sender
        assert delivered_envelope.protocol_id == envelope.protocol_id
        assert delivered_envelope.message == envelope.message

    def test_envelope_echoed_back(self):
        """Test the envelope is echoed back."""
        addr_1 = self.connection_client_1.address
        addr_2 = self.connection_client_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_2,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_client_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_2

        self.multiplexer_client_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    def test_envelope_echoed_back_node_agent(self):
        """Test the envelope is echoed back node agent."""
        addr_1 = self.connection_client_1.address
        addr_n = self.connection_node_2.address

        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )
        original_envelope = Envelope(
            to=addr_n,
            sender=addr_1,
            protocol_id=DefaultMessage.protocol_id,
            message=DefaultSerializer().encode(msg),
        )

        self.multiplexer_client_1.put(original_envelope)
        delivered_envelope = self.multiplexer_node_2.get(block=True, timeout=10)
        assert delivered_envelope is not None

        delivered_envelope.to = addr_1
        delivered_envelope.sender = addr_n

        self.multiplexer_node_2.put(delivered_envelope)
        echoed_envelope = self.multiplexer_client_1.get(block=True, timeout=5)

        assert echoed_envelope is not None
        assert echoed_envelope.to == original_envelope.sender
        assert delivered_envelope.sender == original_envelope.to
        assert delivered_envelope.protocol_id == original_envelope.protocol_id
        assert delivered_envelope.message == original_envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        for mux in reversed(cls.mutliplexers):
            mux.disconnect()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


@libp2p_log_on_failure_all
class TestLibp2pClientConnectionRouting:
    """Test that libp2p DHT network will reliably route envelopes from clients connected to different nodes"""

    @classmethod
    @libp2p_log_on_failure
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        cls.log_files = []
        cls.multiplexers = []

        try:
            cls.connection_node_1 = _make_libp2p_connection(
                port=DEFAULT_PORT + 1,
                delegate_port=DEFAULT_DELEGATE_PORT + 1,
                delegate=True,
            )
            cls.multiplexer_node_1 = Multiplexer([cls.connection_node_1])
            cls.log_files.append(cls.connection_node_1.node.log_file)
            cls.multiplexer_node_1.connect()
            cls.multiplexers.append(cls.multiplexer_node_1)

            entry_peer = cls.connection_node_1.node.multiaddrs[0]

            cls.connection_node_2 = _make_libp2p_connection(
                port=DEFAULT_PORT + 2,
                delegate_port=DEFAULT_DELEGATE_PORT + 2,
                entry_peers=[entry_peer],
                delegate=True,
            )
            cls.multiplexer_node_2 = Multiplexer([cls.connection_node_2])
            cls.log_files.append(cls.connection_node_2.node.log_file)
            cls.multiplexer_node_2.connect()
            cls.multiplexers.append(cls.multiplexer_node_2)

            cls.connections = [cls.connection_node_1, cls.connection_node_2]
            cls.addresses = [
                cls.connection_node_1.address,
                cls.connection_node_2.address,
            ]

            for _ in range(DEFAULT_CLIENTS_PER_NODE):
                for port in [DEFAULT_DELEGATE_PORT + 1, DEFAULT_DELEGATE_PORT + 2]:
                    conn = _make_libp2p_client_connection(port)
                    mux = Multiplexer([conn])

                    cls.connections.append(conn)
                    cls.addresses.append(conn.address)

                    mux.connect()
                    cls.multiplexers.append(mux)
        except Exception:
            cls.teardown_class()
            raise

    def test_connection_is_established(self):
        """Test connection is established."""
        for conn in self.connections:
            assert conn.is_connected is True

    def test_star_routing_connectivity(self):
        """Test routing with star connectivity."""
        msg = DefaultMessage(
            dialogue_reference=("", ""),
            message_id=1,
            target=0,
            performative=DefaultMessage.Performative.BYTES,
            content=b"hello",
        )

        for source in range(len(self.multiplexers)):
            for destination in range(len(self.multiplexers)):
                if destination == source:
                    continue
                envelope = Envelope(
                    to=self.addresses[destination],
                    sender=self.addresses[source],
                    protocol_id=DefaultMessage.protocol_id,
                    message=DefaultSerializer().encode(msg),
                )

                self.multiplexers[source].put(envelope)
                delivered_envelope = self.multiplexers[destination].get(
                    block=True, timeout=10
                )

                assert delivered_envelope is not None
                assert delivered_envelope.to == envelope.to
                assert delivered_envelope.sender == envelope.sender
                assert delivered_envelope.protocol_id == envelope.protocol_id
                assert delivered_envelope.message == envelope.message

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        for mux in reversed(cls.multiplexers):
            mux.disconnect()
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


def test_libp2pclientconnection_uri():
    """Test the uri."""
    uri = Uri(host="127.0.0.1")
    uri = Uri(host="127.0.0.1", port=10000)
    assert uri.host == "127.0.0.1" and uri.port == 10000



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the `aea gui` sub-commands."""

import unittest.mock

from flask import Flask

import aea.cli_gui

from tests.common.mocks import ctx_mock_Popen
from tests.test_cli_gui.test_base import create_app


def test_home_page_exits():
    """Test that the home-page exits."""
    app = create_app()

    # sends HTTP GET request to the application
    # on the specified path
    result = app.get("/")

    # assert the status code of the response
    assert result.status_code == 200
    assert "AEA GUI" in str(result.data)


def test_icon():
    """Test that the home-page exits."""
    app = create_app()

    # sends HTTP GET request to the application
    # on the specified path
    result = app.get("/favicon.ico")

    # assert the status code of the response
    assert result.status_code == 200


def test_js():
    """Test that the home-page exits."""
    app = create_app()

    # sends HTTP GET request to the application
    # on the specified path
    result = app.get("/static/js/home.js")

    # assert the status code of the response
    assert result.status_code == 200


def test_run_app():
    """Test that running the app in non-test mode works."""
    with ctx_mock_Popen():
        with unittest.mock.patch.object(Flask, "run", return_value=None):
            aea.cli_gui.run(8080)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the gym connection module."""
import asyncio
import logging
import os
from typing import cast
from unittest.mock import patch

import gym
import pytest

from aea.common import Address
from aea.configurations.base import ConnectionConfig
from aea.identity.base import Identity
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.fetchai.connections.gym.connection import GymConnection
from packages.fetchai.protocols.gym.dialogues import GymDialogue
from packages.fetchai.protocols.gym.dialogues import GymDialogues as BaseGymDialogues
from packages.fetchai.protocols.gym.message import GymMessage

from tests.conftest import ROOT_DIR, UNKNOWN_PROTOCOL_PUBLIC_ID


logger = logging.getLogger(__name__)


class GymDialogues(BaseGymDialogues):
    """The dialogues class keeps track of all gym dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return GymDialogue.Role.AGENT

        BaseGymDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


class TestGymConnection:
    """Test the packages/connection/gym/connection.py."""

    def setup(self):
        """Initialise the class."""
        self.env = gym.GoalEnv()
        configuration = ConnectionConfig(connection_id=GymConnection.connection_id)
        self.agent_address = "my_address"
        identity = Identity("name", address=self.agent_address)
        self.gym_con = GymConnection(
            gym_env=self.env, identity=identity, configuration=configuration
        )
        self.loop = asyncio.get_event_loop()
        self.gym_address = str(GymConnection.connection_id)
        self.dialogues = GymDialogues(self.agent_address)

    def teardown(self):
        """Clean up after tests."""
        self.loop.run_until_complete(self.gym_con.disconnect())

    @pytest.mark.asyncio
    async def test_gym_connection_connect(self):
        """Test the connection None return value after connect()."""
        assert self.gym_con.channel._queue is None
        await self.gym_con.channel.connect()
        assert self.gym_con.channel._queue is not None

    @pytest.mark.asyncio
    async def test_decode_envelope_error(self):
        """Test the decoding error for the envelopes."""
        await self.gym_con.connect()
        envelope = Envelope(
            to=self.gym_address,
            sender=self.agent_address,
            protocol_id=UNKNOWN_PROTOCOL_PUBLIC_ID,
            message=b"hello",
        )

        with pytest.raises(ValueError):
            await self.gym_con.send(envelope)

    @pytest.mark.asyncio
    async def test_send_connection_error(self):
        """Test send connection error."""
        msg, sending_dialogue = self.dialogues.create(
            counterparty=self.gym_address, performative=GymMessage.Performative.RESET,
        )
        envelope = Envelope(
            to=msg.to, sender=msg.sender, protocol_id=msg.protocol_id, message=msg,
        )

        with pytest.raises(ConnectionError):
            await self.gym_con.send(envelope)

    @pytest.mark.asyncio
    async def test_send_act(self):
        """Test send act message."""
        sending_dialogue = await self.send_reset()
        assert sending_dialogue.last_message is not None
        msg = sending_dialogue.reply(
            performative=GymMessage.Performative.ACT,
            action=GymMessage.AnyObject("any_action"),
            step_id=1,
        )
        envelope = Envelope(
            to=msg.to, sender=msg.sender, protocol_id=msg.protocol_id, message=msg,
        )
        await self.gym_con.connect()

        observation = 1
        reward = 1.0
        done = True
        info = "some info"
        with patch.object(
            self.env, "step", return_value=(observation, reward, done, info)
        ) as mock:
            await self.gym_con.send(envelope)
            mock.assert_called()

        response = await asyncio.wait_for(self.gym_con.receive(), timeout=3)
        response_msg = cast(GymMessage, response.message)
        response_dialogue = self.dialogues.update(response_msg)

        assert response_msg.performative == GymMessage.Performative.PERCEPT
        assert response_msg.step_id == msg.step_id
        assert response_msg.observation.any == observation
        assert response_msg.reward == reward
        assert response_msg.done == done
        assert response_msg.info.any == info
        assert sending_dialogue == response_dialogue

    @pytest.mark.asyncio
    async def test_send_reset(self):
        """Test send reset message."""
        _ = await self.send_reset()

    @pytest.mark.asyncio
    async def test_send_close(self):
        """Test send close message."""
        sending_dialogue = await self.send_reset()
        assert sending_dialogue.last_message is not None
        msg = sending_dialogue.reply(performative=GymMessage.Performative.CLOSE,)
        envelope = Envelope(
            to=msg.to, sender=msg.sender, protocol_id=msg.protocol_id, message=msg,
        )
        await self.gym_con.connect()

        with patch.object(self.env, "close") as mock:
            await self.gym_con.send(envelope)
            mock.assert_called()

    @pytest.mark.asyncio
    async def test_send_close_negative(self):
        """Test send close message with invalid reference and message id and target."""
        incorrect_msg = GymMessage(
            performative=GymMessage.Performative.CLOSE,
            dialogue_reference=self.dialogues.new_self_initiated_dialogue_reference(),
        )
        incorrect_msg.to = self.gym_address
        incorrect_msg.sender = self.agent_address

        # the incorrect message cannot be sent into a dialogue, so this is omitted.

        envelope = Envelope(
            to=incorrect_msg.to,
            sender=incorrect_msg.sender,
            protocol_id=incorrect_msg.protocol_id,
            message=incorrect_msg,
        )
        await self.gym_con.connect()

        with patch.object(self.gym_con.channel.logger, "warning") as mock_logger:
            await self.gym_con.send(envelope)
            mock_logger.assert_any_call(
                f"Could not create dialogue from message={incorrect_msg}"
            )

    async def send_reset(self) -> GymDialogue:
        """Send a reset."""
        msg, sending_dialogue = self.dialogues.create(
            counterparty=self.gym_address, performative=GymMessage.Performative.RESET,
        )
        assert sending_dialogue is not None
        envelope = Envelope(
            to=msg.to, sender=msg.sender, protocol_id=msg.protocol_id, message=msg,
        )
        await self.gym_con.connect()

        with patch.object(self.env, "reset") as mock:
            await self.gym_con.send(envelope)
            mock.assert_called()

        response = await asyncio.wait_for(self.gym_con.receive(), timeout=3)
        response_msg = cast(GymMessage, response.message)
        response_dialogue = self.dialogues.update(response_msg)

        assert response_msg.performative == GymMessage.Performative.STATUS
        assert response_msg.content == {"reset": "success"}
        assert sending_dialogue == response_dialogue
        return sending_dialogue

    @pytest.mark.asyncio
    async def test_receive_connection_error(self):
        """Test receive connection error and Cancel Error."""
        with pytest.raises(ConnectionError):
            await self.gym_con.receive()

    def test_gym_env_load(self):
        """Load gym env from file."""
        curdir = os.getcwd()
        os.chdir(os.path.join(ROOT_DIR, "examples", "gym_ex"))
        gym_env_path = "gyms.env.BanditNArmedRandom"
        configuration = ConnectionConfig(
            connection_id=GymConnection.connection_id, env=gym_env_path
        )
        identity = Identity("name", address=self.agent_address)
        gym_con = GymConnection(
            gym_env=None, identity=identity, configuration=configuration
        )
        assert gym_con.channel.gym_env is not None
        os.chdir(curdir)
